!
! Object I/O module
!
! 2012 Musa Maharramov
! Stanford Exploration Project
!
! type(persistent) object
! implementing a persistent
! data structure of arbitrary
! complexity
!
module io
use iso_c_binding
use error
use IFPORT
implicit none
!
integer, parameter :: SIZEOFINT = 4
integer, parameter :: SIZEOFREAL = 4
!
integer, parameter :: IO_STRLEN=4096
! max number of members in a persistent type
integer, parameter :: IO_MAX_MEMBER=1024

! max file unit
integer, private :: maxunit = 500

public :: getarg

! member of a persistent type
type :: member_type
! the actual member entry
  character(len=IO_STRLEN) ::  entry
  ! member name
  character(len=IO_STRLEN) :: sname
! member type  
  character(len=IO_STRLEN) :: stype
  character(len=IO_STRLEN) :: str
  integer :: ival
  real ::  fval
! member no of dimensions
  integer :: ndim
! dimensions
  integer, dimension(1:20) :: dims
end type member_type

! header for a persistent type
type, public :: header_type

! no of members
integer ::  n
! optional binary file handle
integer :: hbin = 0
! header members
class(member_type), dimension(:), pointer :: member
! header file descriptor
integer :: hd
! open for writing or reading
logical :: bwrite

contains
  procedure, public, pass(this) :: in
  procedure, public, pass(this) :: out
  procedure, public, pass(this) :: defined
  procedure, public, pass(this) :: parse
  procedure, public, pass(this) :: get_member    
  procedure, public, pass(this) :: close
! members loading
  procedure, public, pass(this) :: loads
  procedure, public, pass(this) :: loadi
  procedure, public, pass(this) :: loadc
  procedure, public, pass(this) :: loadac
  procedure, public, pass(this) :: loadia1
  procedure, public, pass(this) :: loadia2
  procedure, public, pass(this) :: loadia3
  procedure, public, pass(this) :: loadf
  procedure, public, pass(this) :: loadfa1
  procedure, public, pass(this) :: loadfa2
  procedure, public, pass(this) :: loadfa3
  generic, public :: load => loads,  &
    loadi,loadc,loadac,loadia1,loadia2, &
    loadia3, loadf,loadfa1,loadfa2,loadfa3
! member storing
  procedure, public, pass(this) :: stores
  procedure, public, pass(this) :: storei
  procedure, public, pass(this) :: storec
  procedure, public, pass(this) :: storeac
  procedure, public, pass(this) :: storeia1
  procedure, public, pass(this) :: storeia2
  procedure, public, pass(this) :: storeia3
  procedure, public, pass(this) :: storef
  procedure, public, pass(this) :: storefa1
  procedure, public, pass(this) :: storefa2
  procedure, public, pass(this) :: storefa3
  generic, public :: store => stores, &
  storei,storec,storeac,storeia1,storeia2, &
  storeia3, storef,storefa1,storefa2,storefa3
end type header_type

! abstract persistent type
type, abstract, public :: persistent
contains
      procedure(load_), deferred :: load
      procedure(store_), deferred :: store 
end type persistent 

abstract interface 
subroutine load_(this, header) 
import :: persistent, header_type
class(persistent), intent(inout) :: this
class(header_type), intent(inout)::  header
end subroutine load_

subroutine store_(this, header) 
import :: persistent, header_type
class(persistent), intent(inout) :: this
class(header_type), intent(inout)::  header
end subroutine store_
end interface 

! POSIX system call: open file/pipe for reading
interface
  function posixopenr(s) bind(c)
  use iso_c_binding
  character(kind=c_char) :: s(*)
  integer(kind=c_int) :: posixopenr
  end function posixopenr
end interface

! POSIX system call: open file/pipe for writing
interface
  function posixopenw(s) bind(c)
  use iso_c_binding
  character(kind=c_char) :: s(*)
  integer(kind=c_int) :: posixopenw
  end function posixopenw
end interface

! POSIX system call: close handle
interface
  integer function posixclose(h) bind(c)
  use iso_c_binding
  integer(kind=c_int), intent(in), value :: h
  end function posixclose
end interface

contains

! get command line argument 'sname'=
character(len=IO_STRLEN) function getarg(sname,defval)
character(len=*), intent(in) :: sname
character(len=*), optional, intent(in) :: defval 
! aux
integer :: narg, i
character(len=IO_STRLEN) :: cmd
  getarg=''
  narg=command_argument_count()
  if (narg==0) then
    if (present(defval)) getarg=defval
  else
    do i=1,narg
      call get_command_argument(i,cmd)
      cmd=adjustl(cmd)
      if (index(cmd,sname)==1) then
        cmd=adjustl(cmd(len_trim(sname)+1:))
        if (index(cmd,'=')==1) then
          getarg=trim(adjustl(cmd(2:)))
          return
        end if
      end if
    end do
  end if
end function getarg

! expand an anetry's the assigned value 
! to a full path unless it looks like
! a full path
subroutine expandpath(entry,defpath)
character(len=*), intent(inout) :: entry
character(len=*), intent(in), optional :: defpath
! aux
character(len=IO_STRLEN) ::temppath
integer :: ier, j
  j=index(entry,'=')
  if (j==0) return
  if (present(defpath)) then
    temppath=defpath
  else
    ier=getcwd(temppath)
  end if
  if (ier/=0) then
     call set_error(IO_GETCWD_ER)
     return
  end if
  entry=entry(1:j)//trim(temppath)//'/'//& 
    trim(adjustl(entry(j+1:)))
end subroutine expandpath

function isint(s)
character(len=*),intent(in) :: s
logical :: isint
  if (trim(s)==trim('integer')) then
    isint=.true.
  else
    isint=.false.
  end if
end function isint

function isreal(s)
character(len=*),intent(in) :: s
logical :: isreal
  if (trim(s)==trim('real')) then
    isreal=.true.
  else
    isreal=.false.
  end if
end function isreal

function isstr(s)
character(len=*),intent(in) :: s
logical :: isstr
  if (trim(s)==trim('character*')) then
    isstr=.true.
  else
    isstr=.false.
  end if
end function isstr

! copy Fortran string to a char array
subroutine copy2chara(s1, s2)
character(len=*), intent(in) :: s1
character(kind=c_char) :: s2(*)
integer :: i,n
  n=min(IO_STRLEN-1,len_trim(s1))
  do i=1,n
    s2(i)=s1(i:i)
  end do
  s2(n+1)=achar(0)
end subroutine copy2chara

! open header with the given name
! and optional binary
subroutine in(this, header, binary)

class(header_type), intent(inout) :: this
character(len=*), intent(in) :: header
character(len=*), optional, intent(in) :: binary
! aux
integer :: ier
character(len=IO_STRLEN) :: stemp
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: narg, i
  if (associated(this%member)) deallocate(this%member)
  this%n=0
  allocate(this%member(1:IO_MAX_MEMBER))
  if (len_trim(header)/=0) then ! file header
!$omp critical  
    maxunit=maxunit+1
!$omp end critical  
    this%hd=maxunit
    open (this%hd,file=header, action='read',iostat=ier)
    if (ier>0) then
      call set_error(IO_HEADER_OPEN_ER)
      return
    end if
    this%bwrite=.false.
    READHDR: do    
      read (this%hd,'(A)',iostat=ier) stemp
      if (ier>0) then
        call set_error(IO_HEADER_READ_ER)
        return
      else if (ier<0) then
        exit
      end if
      if (len_trim(stemp)==0) cycle
      this%n=this%n+1
!     init new member record
      this%member(this%n)%stype=''
      this%member(this%n)%sname=''
      this%member(this%n)%str=''
      this%member(this%n)%ival=0
      this%member(this%n)%fval=0.
      this%member(this%n)%ndim=0
      stemp=adjustl(stemp)
      this%member(this%n)%entry=stemp
    end do READHDR
  else ! command-line header
    narg=command_argument_count()
    if (narg==0) then
      call set_error(IO_HEADER_OPEN_ER)
      return
    else
      do i=1,narg
        this%n=this%n+1
        call get_command_argument(i, &
          this%member(this%n)%entry)
!       init new member record
        this%member(this%n)%stype=''
        this%member(this%n)%sname=''
        this%member(this%n)%str=''
        this%member(this%n)%ival=0
        this%member(this%n)%fval=0.
        this%member(this%n)%ndim=0
      end do
    end if
  end if
!
! open the associated binary
  if (present(binary)) then
    call copy2chara(binary,tempchara)
    this%hbin=posixopenr(tempchara);
    if (this%hbin<=0) then
      call set_error(IO_BIN_OPEN_ER)    
      return
    end if
  else
    this%hbin=0
  end if

  call this%parse()
end subroutine in 

! (re)parse and (re)compile header members
subroutine parse(this)
use iso_c_binding
class(header_type),intent(inout) :: this
character(len=IO_STRLEN) :: stemp,name1

integer :: n,i,j,k,ier

  READHDR: do n=1,this%n
    stemp=this%member(n)%entry
!   first token right delimiter -- space, = or (
    j=scan(stemp,'=( ')
    if (j<=1) then
      call set_error(IO_HDR_SYNTAX_ER)
      return
    end if

    name1=stemp(1:j-1)
    stemp=adjustl(stemp(j:))
    j=scan(stemp,'(=')
    if (j==1) then
!     unspecified type
      this%member(n)%stype=''
    else
      this%member(n)%stype=name1
      name1=stemp(1:j-1)
      stemp=adjustl(stemp(j:))
    end if
!
    this%member(n)%sname=name1
!   by now both type and name have been set
    this%member(n)%ndim=0
    ARRAY: if (index(stemp,'(')==1) then
      if (index(stemp,')')==0) then
        call set_error(IO_HDR_SYNTAX_ER)
        return
      end if
      stemp=adjustl(stemp(2:))
      GETDIMS: do
        this%member(n)%ndim= &
          this%member(n)%ndim+1
        k=scan(stemp,',)')
        if (k==1) then
          call set_error(IO_HDR_SYNTAX_ER)
          return
        end if
        read(stemp(1:k-1),*) name1
        j=this%get_member(name1)
        if (j<=0) then
          read(stemp(1:k-1),*,iostat=ier) j
          if (j<=0 .or. ier>0) then
            call set_error(IO_HDR_SYNTAX_ER)
            return
          end if
          this%member(n)%dims(this%member(n)%ndim)=j
        else if (.not.isint(this%member(j)%stype)) then
          call set_error(IO_BAD_AXIS_ER)
          return
!       if reading values from binaries, allow dynamic axis sizes
        else if (this%hbin/=0) then
          this%member(n)%dims(this%member(n)%ndim)= -j
!       statically compute axis sizes from header initialisation
!       expressions specified so far
        else
          this%member(n)%dims(this%member(n)%ndim)= &
            this%member(j)%ival
          if (this%member(j)%ival <= 0) then
            call set_error(IO_BAD_AXIS_ER)
            return
          end if 
        end if
        stemp=adjustl(stemp(k:))
        if (stemp(1:1)==')') exit
        stemp=adjustl(stemp(2:))
      end do GETDIMS      
    end if ARRAY
!
!   by now dims have been set, now determine the value
    i=index(stemp,'=');
    INIT: if (i==0) then
!     a binary is required
      if (this%hbin==0) then
        call set_error(IO_NOVAL_ER)
        return
      end if
    else 
!     explicit initialisation present
      stemp=adjustl(stemp(i+1:))
      this%member(n)%str=trim(stemp)
!     read if scalar
      RDSCALAR: if (this%member(n)%ndim==0) then
        if (isint(this%member(n)%stype)) then
          read (stemp,'(i)') this%member(n)%ival
        else if (isreal(this%member(n)%stype)) then
           read (stemp,'(f)') this%member(n)%fval
        else if (isstr(this%member(n)%stype)) then
           this%member(n)%str=stemp
!       try to figure out the type if unspecified
        else if (len_trim(this%member(n)%stype)==0) then
          if (verify(trim(stemp),'-+0123456789 ')==0) then ! int
            read (stemp,'(i)') this%member(n)%ival
            this%member(n)%stype='integer'
          else if (verify(trim(stemp),'-+0123456789.eE ')==0) then ! real
            read (stemp,'(f)') this%member(n)%fval
            this%member(n)%stype='real'
          else 
            call set_error(IO_UNTYPED_MEMBER_ER)
            call warn('offending value='//trim(stemp))
            return
          end if
        end if
      end if RDSCALAR
    end if INIT
!   arrays of strings not supported yet    
    if (isstr(this%member(n)%stype) .and. this%member(n)%ndim >0) then
      call set_error(IO_HDR_SYNTAX_ER)
      return
    end if
  end do READHDR
! 
end subroutine parse 

! create header with the given name
! and optional binary
subroutine out(this, header, binary)
class(header_type), intent(inout) :: this
character(len=*), intent(in) :: header
character(len=*), optional, intent(in) :: binary
integer :: ier
  if (associated(this%member)) deallocate(this%member)
  this%n=0
  allocate(this%member(1:IO_MAX_MEMBER))
!$omp critical  
  maxunit=maxunit+1
!$omp end critical 
  this%hd=maxunit
  open (this%hd,file=header, action='write',iostat=ier)
  if (ier>0) then
    call set_error(IO_HEADER_OPEN_ER)
    return
  end if
  this%bwrite=.true.
end subroutine out


! close the header/commit the changes
subroutine close(this)
class(header_type), intent(inout) :: this
! aux
integer :: i, ier
character(len=IO_STRLEN) :: formatstr
! write the header to disk if in write mode
  if (this%bwrite) then
!   output format to accommodate the entire entry
    write (formatstr,*) IO_STRLEN
    formatstr='('//trim(adjustl(formatstr))//'A)'
    do i=1,this%n
      write (this%hd,trim(formatstr),iostat=ier)  &
       trim(this%member(i)%entry)
      if (ier>0) then
        call set_error(IO_HEADER_WRITE_ER)
        return
      end if
    end do
  end if
  close(this%hd)
! close the associated binary
  if (this%hbin>0) then
    if (posixclose(this%hbin)<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if
  end if
end subroutine close

! return the index of the  member with name s
! or 0
function get_member(this, s)
class(header_type), intent(inout) :: this
character(len=*), intent(in) :: s
integer :: get_member
! aux
integer :: i
! allow later additions to override 
! previous declarations
  do i=this%n,1,-1
    if (trim(s)==trim(this%member(i)%sname)) then
      get_member=i
      return
    end if
  end do
  get_member=0
  return
end function get_member


! load int scalar v with name s
! from header h
subroutine loadi(this, v, s, defval)
class(header_type), intent(inout) :: this
integer, intent(inout) :: v
integer, optional, intent(in) :: defval
character(len=*), intent(in) :: s
! aux
integer :: i,j,sizev,cnt
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    if (present(defval)) then
      v=defval
      return
    else
      call set_error(IO_NOSUCHMEMBER_ER)
      return
    end if
  end if
  if (.not.isint(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim /=0) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
! scalar size
  sizev=SIZEOFINT
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
    read (this%member(j)%str,'(i)') v
    call set_error(IO_OK)
  else
!   call C posixered file read function
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
  this%member(j)%ival=v
end subroutine loadi

! load cb bytes into a variable v of any type
! with name s, from header h
subroutine loadc(this, v, cb, s)
class(header_type), intent(inout) :: this
class(*), intent(inout) :: v
integer, intent(in) :: cb
character(len=*), intent(in) :: s
! aux
integer :: i,j,k,cnt
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    call set_error(IO_NOSUCHMEMBER_ER)
    return
  end if
!  
  cnt=cb
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
!   C function to open a binary for reading
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenr(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    k=posixread(i, v, cnt)
    if (k/=cnt) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixeread function
    k=posixread(this%hbin, v, cnt)
    if (k/=cnt) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
end subroutine loadc

! load cb bytes into an array v of any type
! with name s, from header h
subroutine loadac(this, v, cb, s)
class(header_type), intent(inout) :: this
class(*),dimension(:), intent(inout) :: v
integer, intent(in) :: cb
character(len=*), intent(in) :: s
! aux
integer :: i,j,k,cnt
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    call set_error(IO_NOSUCHMEMBER_ER)
    return
  end if
  if (this%member(j)%ndim <1) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
  cnt=cb
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
!   C function to open a binary for reading
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenr(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    k=posixread(i, v, cnt)
    if (k/=cnt) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixeread function
    k=posixread(this%hbin, v, cnt)
    if (k/=cnt) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
end subroutine loadac

! load int array v with name s
! from header h
subroutine loadia1(this, v, s)
class(header_type), intent(inout) :: this
integer,dimension(:), intent(inout) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,k,cnt,sizev
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    call set_error(IO_NOSUCHMEMBER_ER)
    return
  end if
  if (.not.isint(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim /=1) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
! array size
  sizev=size(v,1)*SIZEOFINT
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
!   C function to open a binary for reading
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenr(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixread(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixeread function
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
end subroutine loadia1

! load int 2D array v with name s
! from header h
subroutine loadia2(this, v, s)
class(header_type), intent(inout) :: this
integer,dimension(:,:), intent(inout) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,k,cnt,sizev
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    call set_error(IO_NOSUCHMEMBER_ER)
    return
  end if
  if (.not.isint(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim /=2) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
! array size
  sizev=size(v,1)*size(v,2)*SIZEOFINT
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)*size(v,2)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
!   C function to open a binary for reading
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenr(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixread(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixeread function
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
end subroutine loadia2

! load int 3D array v with name s
! from header h
subroutine loadia3(this, v, s)
class(header_type), intent(inout) :: this
integer,dimension(:,:,:), intent(inout) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,k,cnt,sizev
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    call set_error(IO_NOSUCHMEMBER_ER)
    return
  end if
  if (.not.isint(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim /=3) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
! array size
  sizev=size(v,1)*size(v,2)*size(v,3)*SIZEOFINT
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)*size(v,2)*size(v,3)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
!   C function to open a binary for reading
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenr(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixread(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixeread function
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
end subroutine loadia3


! load real scalar v with name s
! from header h
subroutine loadf(this, v, s, defval)
class(header_type), intent(inout) :: this
real, intent(inout) :: v
real, optional, intent(in) :: defval
character(len=*), intent(in) :: s
! aux
integer :: i,j,sizev,cnt
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    if (present(defval)) then
      v=defval
      return
    else
      call set_error(IO_NOSUCHMEMBER_ER)
      return
    end if
  end if
  if (.not.isreal(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim /=0) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
! scalar size
  sizev=SIZEOFREAL
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
    read (this%member(j)%str,'(f)') v
    call set_error(IO_OK)
  else
!   call C posixered file read function
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
  this%member(j)%ival=v
end subroutine loadf

! load real 1D array v with name s
! from header h
subroutine loadfa1(this, v, s)
class(header_type), intent(inout) :: this
real,dimension(:), intent(inout) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,k,cnt,sizev
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    call set_error(IO_NOSUCHMEMBER_ER)
    return
  end if
  if (.not.isreal(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
! allow loading multi-dim binary into a 1d array...
!  if (this%member(j)%ndim /=1) then
!    call set_error(IO_DIM_MISMATCH_ER)
!    return
!  end if
!
! ...but always check if array sizes agree
  sizev=size(v,1)*SIZEOFREAL
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
!   C function to open a binary for reading
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenr(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixread(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixeread function
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
end subroutine loadfa1

! load real 2D array v with name s
! from header h
subroutine loadfa2(this, v, s)
class(header_type), intent(inout) :: this
real,dimension(:,:), intent(inout) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,k,cnt,sizev
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    call set_error(IO_NOSUCHMEMBER_ER)
    return
  end if
  if (.not.isreal(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim /=2) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
! array size
  sizev=size(v,1)*size(v,2)*SIZEOFREAL
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)*size(v,2)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
!   C function to open a binary for reading
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenr(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixread(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixeread function
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
end subroutine loadfa2

! load real 3D array v with name s
! from header h
subroutine loadfa3(this, v, s)
class(header_type), intent(inout) :: this
real,dimension(:,:,:), intent(inout) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,k,cnt,sizev
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    call set_error(IO_NOSUCHMEMBER_ER)
    return
  end if
  if (.not.isreal(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim /=3) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
! array size
  sizev=size(v,1)*size(v,2)*size(v,3)*SIZEOFREAL
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)*size(v,2)*size(v,3)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
!   C function to open a binary for reading
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenr(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixread(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixeread function
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
end subroutine loadfa3

! store int scalar v with name s
! from header h
subroutine storei(this, v, s)
class(header_type), intent(inout) :: this
integer, intent(in) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,sizev,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'integer ')==0)  then
    this%member(this%n)%entry= &
      'integer '//this%member(this%n)%entry
  end if

! scalar size
  sizev=SIZEOFINT
!
! if no binary specified, write value to
! the header itself
  j=this%n
  if (this%hbin<=0) then
    write (this%member(j)%str,*) v
    i=index(s,'=');
    if (i==0) then ! append assignment to the entry
       this%member(this%n)%entry= &
         trim(this%member(this%n)%entry)//'=';
         i=index(this%member(this%n)%entry,'=');
    else
      this%member(this%n)%entry(i+1:)=''
    end if
!   now append the value to the entry
    write (this%member(j)%entry(i+1:),'(i)') v
    call set_error(IO_OK)
  else
!   call C posixwrite 
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
! 
!  

! reparse to check for errors
  call this%parse()
!
  j=this%n
  if (j==0) return
  if (.not.isint(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim/=0) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
end subroutine storei

! store cb bytes of a var v
! of any type with name s, from header h
subroutine storec(this, v, cb, s)
class(header_type), intent(inout) :: this
class(*), intent(in) :: v
integer, intent(in) :: cb
character(len=*), intent(in) :: s
! for string manipulation
character(len=IO_STRLEN) :: rightstr
! POSIX IO names
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: i,j,k,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
!
! parse the header
  call this%parse()
!
  j=this%n
  if (this%member(j)%ndim < 1) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!  
  cnt=cb
!
! if no binary specified, write values 
! to stream (binary) specified after '='
  j=this%n
  if (this%hbin<=0) then
!   verify if a path has been specified,
!   and prepend to filename if not
    call expandpath(this%member(j)%entry)
!   C function to open a binary for writing
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenw(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    k=posixwrite(i, v, cnt)
    if (k/=cnt) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixwrite function
    k=posixwrite(this%hbin, v, cnt)
    if (k/=cnt) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
!  
end subroutine storec

! store cb bytes of 1D array v
! of any type with name s, from header h
subroutine storeac(this, v, cb, s)
class(header_type), intent(inout) :: this
class(*), dimension(:), intent(in) :: v
integer, intent(in) :: cb
character(len=*), intent(in) :: s
! for string manipulation
character(len=IO_STRLEN) :: rightstr
! POSIX IO names
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: i,j,k,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
!
! parse the header
  call this%parse()
!
  j=this%n
  if (this%member(j)%ndim < 1) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!  
  cnt=cb
!
! if no binary specified, write values 
! to stream (binary) specified after '='
  j=this%n
  if (this%hbin<=0) then
!   verify if a path has been specified,
!   and prepend to filename if not
    call expandpath(this%member(j)%entry)
!   C function to open a binary for writing
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenw(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    k=posixwrite(i, v, cnt)
    if (k/=cnt) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixwrite function
    k=posixwrite(this%hbin, v, cnt)
    if (k/=cnt) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
!  
end subroutine storeac

! store int 1D array v with name s
! from header h
subroutine storeia1(this, v, s)
class(header_type), intent(inout) :: this
integer, dimension(:), intent(in) :: v
character(len=*), intent(in) :: s
! for string manipulation
character(len=IO_STRLEN) :: rightstr
! POSIX IO names
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: i,j,k,sizev,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'integer ')==0)  then
    this%member(this%n)%entry= &
      'integer '//this%member(this%n)%entry
  end if
! append array shape and size if necessary
  if (index(this%member(this%n)%entry,'(')==0)  then
    i=index(this%member(this%n)%entry,'=')
    if (i ==0) then
      rightstr=''
    else
      rightstr=this%member(this%n)%entry(i:)
      this%member(this%n)%entry(i:)='';
    end if
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//'('
    write(this%member(this%n)%entry(i+1:),*) size(v,1)
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//')'
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//rightstr
  end if
!
! parse the header
  call this%parse()
!
  if (.not.isint(this%member(this%n)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(this%n)%sname,this%n)
    return
  end if
!
  j=this%n
  if (this%member(j)%ndim /= 1) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
! array size
  sizev=size(v,1)*SIZEOFINT
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, write values 
! to stream (binary) specified after '='
  j=this%n
  if (this%hbin<=0) then
!   verify if a path has been specified,
!   and prepend to filename if not
    call expandpath(this%member(j)%entry)
!   C function to open a binary for writing
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenw(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixwrite(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixwrite function
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
!  
end subroutine storeia1

! store int 2D array v with name s
! from header h
subroutine storeia2(this, v, s)
class(header_type), intent(inout) :: this
integer, dimension(:,:), intent(in) :: v
character(len=*), intent(in) :: s
! for string manipulation
character(len=IO_STRLEN) :: rightstr
! POSIX IO names
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: i,j,k,sizev,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'integer ')==0)  then
    this%member(this%n)%entry= &
      'integer '//this%member(this%n)%entry
  end if
! append array shape and size if necessary
  if (index(this%member(this%n)%entry,'(')==0)  then
    i=index(this%member(this%n)%entry,'=')
    if (i ==0) then
      rightstr=''
    else
      rightstr=this%member(this%n)%entry(i:)
      this%member(this%n)%entry(i:)='';
    end if
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//'('
    write(this%member(this%n)%entry(i+1:),*) size(v,1),',',size(v,2)
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//')'
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//rightstr
  end if
!
! parse the header
  call this%parse()
!
  if (.not.isint(this%member(this%n)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(this%n)%sname,this%n)
    return
  end if
!
  j=this%n
  if (this%member(j)%ndim /= 2) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
! array size
  sizev=size(v,1)*size(v,2)*SIZEOFINT
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)*size(v,2)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, write values 
! to stream (binary) specified after '='
  j=this%n
  if (this%hbin<=0) then
!   verify if a path has been specified,
!   and prepend to filename if not
    call expandpath(this%member(j)%entry)
!   C function to open a binary for writing
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenw(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixwrite(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixwrite function
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
!  
end subroutine storeia2

! store int 3D array v with name s
! from header h
subroutine storeia3(this, v, s)
class(header_type), intent(inout) :: this
integer, dimension(:,:,:), intent(in) :: v
character(len=*), intent(in) :: s
! for string manipulation
character(len=IO_STRLEN) :: rightstr
! POSIX IO names
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: i,j,k,sizev,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'integer ')==0)  then
    this%member(this%n)%entry= &
      'integer '//this%member(this%n)%entry
  end if
! append array shape and size if necessary
  if (index(this%member(this%n)%entry,'(')==0)  then
    i=index(this%member(this%n)%entry,'=')
    if (i ==0) then
      rightstr=''
    else
      rightstr=this%member(this%n)%entry(i:)
      this%member(this%n)%entry(i:)='';
    end if
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//'('
    write(this%member(this%n)%entry(i+1:),*) &
         size(v,1),',',size(v,2),',',size(v,3)
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//')'
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//rightstr
  end if
!
! parse the header
  call this%parse()
!
  if (.not.isint(this%member(this%n)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(this%n)%sname,this%n)
    return
  end if
!
  j=this%n
  if (this%member(j)%ndim /= 3) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
! array size
  sizev=size(v,1)*size(v,2)*size(v,3)*SIZEOFINT
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)*size(v,2)*size(v,3)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, write values 
! to stream (binary) specified after '='
  j=this%n
  if (this%hbin<=0) then
!   verify if a path has been specified,
!   and prepend to filename if not
    call expandpath(this%member(j)%entry)
!   C function to open a binary for writing
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenw(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixwrite(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixwrite function
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
!  
end subroutine storeia3


! store real scalar v with name s
! from header h
subroutine storef(this, v, s)
class(header_type), intent(inout) :: this
real, intent(in) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,sizev,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'real ')==0)  then
    this%member(this%n)%entry= &
      'real '//this%member(this%n)%entry
  end if

! scalar size
  sizev=SIZEOFREAL
!
! if no binary specified, write value to
! the header itself
  j=this%n
  if (this%hbin<=0) then
    write (this%member(j)%str,*) v
    i=index(s,'=');
    if (i==0) then ! append assignment to the entry
       this%member(this%n)%entry= &
         trim(this%member(this%n)%entry)//'=';
         i=index(this%member(this%n)%entry,'=');
    else
      this%member(this%n)%entry(i+1:)=''
    end if
!   now append the value to the entry
    write (this%member(j)%entry(i+1:),'(f)') v
    call set_error(IO_OK)
  else
!   call C posixwrite 
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
! 
!  
! reparse to check for errors
  call this%parse()
!
  j=this%n
  if (j==0) return
  if (.not.isreal(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim/=0) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
end subroutine storef

! store real 1D array v with name s
! from header h
subroutine storefa1(this, v, s)
class(header_type), intent(inout) :: this
real, dimension(:), intent(in) :: v
character(len=*), intent(in) :: s
! for string manipulation
character(len=IO_STRLEN) :: rightstr
! POSIX IO names
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: i,j,k,sizev,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'real ')==0)  then
    this%member(this%n)%entry= &
      'real '//this%member(this%n)%entry
  end if
! append array shape and size if necessary
  if (index(this%member(this%n)%entry,'(')==0)  then
    i=index(this%member(this%n)%entry,'=')
    if (i ==0) then
      rightstr=''
    else
      rightstr=this%member(this%n)%entry(i:)
      this%member(this%n)%entry(i:)='';
    end if
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//'('
    write(this%member(this%n)%entry(i+1:),*) size(v,1)
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//')'
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//rightstr
  end if
!
! parse the header
  call this%parse()
!
  if (.not.isreal(this%member(this%n)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(this%n)%sname,this%n)
    return
  end if
!
  j=this%n
! allow packing a 1D array into a multi-dimensional
! array binary...
!  if (this%member(j)%ndim /= 1) then
!    call set_error(IO_DIM_MISMATCH_ER)
!    return    
!  end if
!
! ...but always check that the sizes agree
  sizev=size(v,1)*SIZEOFREAL
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, write values 
! to stream (binary) specified after '='
  j=this%n
  if (this%hbin<=0) then
!   verify if a path has been specified,
!   and prepend to filename if not
    call expandpath(this%member(j)%entry)
!   C function to open a binary for writing
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenw(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixwrite(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixwrite function
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
!  
end subroutine storefa1

! store int 2D array v with name s
! from header h
subroutine storefa2(this, v, s)
class(header_type), intent(inout) :: this
real, dimension(:,:), intent(in) :: v
character(len=*), intent(in) :: s
! for string manipulation
character(len=IO_STRLEN) :: rightstr
! POSIX IO names
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: i,j,k,sizev,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'real ')==0)  then
    this%member(this%n)%entry= &
      'real '//this%member(this%n)%entry
  end if
! append array shape and size if necessary
  if (index(this%member(this%n)%entry,'(')==0)  then
    i=index(this%member(this%n)%entry,'=')
    if (i ==0) then
      rightstr=''
    else
      rightstr=this%member(this%n)%entry(i:)
      this%member(this%n)%entry(i:)='';
    end if
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//'('
    write(this%member(this%n)%entry(i+1:),*) size(v,1),',',size(v,2)
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//')'
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//rightstr
  end if
!
! parse the header
  call this%parse()
!
  if (.not.isreal(this%member(this%n)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(this%n)%sname,this%n)
    return
  end if
!
  j=this%n
  if (this%member(j)%ndim /= 2) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
! array size
  sizev=size(v,1)*size(v,2)*SIZEOFREAL
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)*size(v,2)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, write values 
! to stream (binary) specified after '='
  j=this%n
  if (this%hbin<=0) then
!   verify if a path has been specified,
!   and prepend to filename if not
    call expandpath(this%member(j)%entry)
!   C function to open a binary for writing
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenw(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixwrite(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixwrite function
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
!  
end subroutine storefa2

! store int 3D array v with name s
! from header h
subroutine storefa3(this, v, s)
class(header_type), intent(inout) :: this
real, dimension(:,:,:), intent(in) :: v
character(len=*), intent(in) :: s
! for string manipulation
character(len=IO_STRLEN) :: rightstr
! POSIX IO names
character(kind=c_char), dimension(1:IO_STRLEN) :: tempchara
! aux
integer :: i,j,k,sizev,cnt
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'real ')==0)  then
    this%member(this%n)%entry= &
      'real '//this%member(this%n)%entry
  end if
! append array shape and size if necessary
  if (index(this%member(this%n)%entry,'(')==0)  then
    i=index(this%member(this%n)%entry,'=')
    if (i ==0) then
      rightstr=''
    else
      rightstr=this%member(this%n)%entry(i:)
      this%member(this%n)%entry(i:)='';
    end if
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//'('
    write(this%member(this%n)%entry(i+1:),*) &
         size(v,1),',',size(v,2),',',size(v,3)
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//')'
    this%member(this%n)%entry= &
    trim(this%member(this%n)%entry)//rightstr
  end if
!
! parse the header
  call this%parse()
!
  if (.not.isreal(this%member(this%n)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(this%n)%sname,this%n)
    return
  end if
!
  j=this%n
  if (this%member(j)%ndim /= 3) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
! array size
  sizev=size(v,1)*size(v,2)*size(v,3)*SIZEOFREAL
  cnt=1
  do i=1,this%member(j)%ndim
    k=this%member(j)%dims(i)
    if (k>=0) then
      cnt=cnt*k
    else
      cnt=cnt*this%member(-k)%ival
    end if
  end do
  if (cnt/=size(v,1)*size(v,2)*size(v,3)) then
    call set_error(IO_DIM_MISMATCH_ER)
    return    
  end if
!
! if no binary specified, write values 
! to stream (binary) specified after '='
  j=this%n
  if (this%hbin<=0) then
!   verify if a path has been specified,
!   and prepend to filename if not
    call expandpath(this%member(j)%entry)
!   C function to open a binary for writing
    call copy2chara(this%member(j)%str,tempchara)
    i=posixopenw(tempchara)
    if (i<=0) then
      call set_error(IO_BIN_OPEN_ER)
      return
    end if
!
    cnt=posixwrite(i, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
    i=posixclose(i)
    if (i<0) then
      call set_error(IO_BIN_CLOSE_ER)
      return
    end if

  else
!   call C posixwrite function
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
!  
end subroutine storefa3

! load string scalar v with name s
! from header h
subroutine loads(this, v, s, defval)
class(header_type), intent(inout) :: this
character(len=*), intent(inout) :: v
character(len=*), optional, intent(in) :: defval
character(len=*), intent(in) :: s
! aux
integer :: i,j,sizev,cnt
! external POSIX call
integer :: posixread
! get entry type from header
  j=this%get_member(s)
  if (j<=0) then
    if (present(defval)) then
      v=defval
      return
    else
      call set_error(IO_NOSUCHMEMBER_ER)
      return
    end if
  end if
  if (.not.isstr(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim /=0) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
  if (len(v)>IO_STRLEN) then
    call set_error(IO_STR_TOO_LONG_ER)
    return
  end if
!
! scalar size
  sizev=len(v)+SIZEOFINT
!
! if no binary specified, read value from
! from the header itself
  if (this%hbin<=0) then
    v=adjustl(this%member(j)%str)
    call set_error(IO_OK)
  else
!   call C posixread
    cnt=posixread(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_READ_ER)
    else
      call set_error(IO_OK)
    end if
  end if
  this%member(j)%str=v
end subroutine loads

! store string scalar v with name s
! from header h
subroutine stores(this, v, s)
class(header_type), intent(inout) :: this
character(len=*), intent(in) :: v
character(len=*), intent(in) :: s
! aux
integer :: i,j,sizev,cnt
character(len=IO_STRLEN) :: tempstr
! external POSIX call
integer :: posixwrite
! append a new entry to the header
  this%n=this%n+1
  this%member(this%n)%entry=s
! prepend type definition if necessary
  if (index(s,'character* ')==0)  then
    this%member(this%n)%entry= &
      'character* '//this%member(this%n)%entry
  end if
  if (len(v)>IO_STRLEN) then
    call set_error(IO_STR_TOO_LONG_ER)
    return
  end if
!
! scalar size
  sizev=len(v)+SIZEOFINT
!
! if no binary specified, write value to
! the header itself
  j=this%n
  if (this%hbin<=0) then
    write (this%member(j)%str,*) v
    i=index(s,'=');
    if (i==0) then ! append assignment to the entry
       this%member(this%n)%entry= &
         trim(this%member(this%n)%entry)//'=';
         i=index(this%member(this%n)%entry,'=');
    end if
!   now append the value to the entry
    write (tempstr,'(4096A)') v
    this%member(j)%entry=trim(this%member(j)%entry(1:i))//&
      trim(adjustl(tempstr))
    call set_error(IO_OK)
  else
!   call C posixwrite 
    cnt=posixwrite(this%hbin, v, sizev)
    if (cnt/=sizev) then
      call set_error(IO_BIN_WRITE_ER)
    else
      call set_error(IO_OK)
    end if
  end if
! 
!  

! reparse to check for errors
  call this%parse()
!
  j=this%n
  if (j==0) return
  if (.not.isstr(this%member(j)%stype)) then
    call set_error(IO_TYPE_MISMATCH_ER)
    call warni('type mismatch for '//this%member(j)%sname,j)
    return
  end if
  if (this%member(j)%ndim/=0) then
    call set_error(IO_DIM_MISMATCH_ER)
    return
  end if
!
end subroutine stores

! check a definition for 'sname'
! is present in the header
logical function defined(this, sname)
class(header_type), intent(inout) :: this
character(len=*), intent(in) :: sname
  defined=this%get_member(sname) >0
end function defined

end module io
