!
! Matrix functions
!
! 2012 Musa Maharramov, Stanford Exploration Project
!
module func
use mkl95_lapack
use mkl95_blas
use error
implicit none

public :: expm

contains


! The accuracy that can be achieved
! using Pade approximation of the 
! exponent, with degree p numerator
! and degree q denominator
! ('eps' argument to 'expm')
function accy(p,q)
  double precision :: accy
  integer, intent(in) :: p, q
!
  integer :: i
  double precision :: r
!
  r=8./(p+q+1)
!
  do i=1,p
    r=r/(2.*(q+i))
  end do
!  
  do i=1,q
    r=r/(2.*(p+i))
  end do
!
  accy=r
  return
end function accy

! exponential of complex matrix A
!
! computes B=exp{A} approximately, returning
! exp{A+E} where |E|<eps
!
subroutine expm(A, eps, B)
  complex, dimension (:, :), intent(in) :: A
  real, intent(in) :: eps
  complex, dimension (:, :), intent(inout) :: B
!
  integer :: n,i,j,q,ier,sgn
  double precision :: e,c,temp
! auxiliary arrays
  double complex, dimension(:,:), allocatable ::  &
    A1,D,Y,X,Z
! required for BLAS zlane()
  double precision, dimension(:,:), allocatable :: &
   work
!
! blas (F77 interface)
  double precision :: zlange
!
  n=size(A,1)
!
  if (eps<=0. .or. size(A,2)/=n .or.  &
    size(B,1)/=n .or. size(B,2)/=n) then
    call warn('expm invalid args')
    call set_error(FUNC_INVPAR_ER)
    return
  end if
  allocate(A1(n,n),D(n,n),Y(n,n),X(n,n), &
   Z(n,n),work(1,n), stat=ier)
  if (ier>0) then
    call warn('expm alloc failed')
    call set_error(FUNC_MEM_ALLOC_ER)
    return
  end if
!
! minimum Pade approximation of degree 5
  A1=A
  temp=zlange('I',n,n,A1,n,work) ! infinity norm
  j=max(0,1+int(log(temp)/log(2.)))
  do i=1,j
    A1=A1/2
  end do
! determine the degree
  q=1
  do 
    if (accy(q,q)<=eps) exit
    q=q+1
  end do
!
  D=0;Y=0;X=0;Z=0;c=1.
  do i=1,n
    D(i,i)=1;Y(i,i)=1;X(i,i)=1 
  end do
!
  sgn=-1
  do i=1,q
    c=c*(q-i+1)/((2.*q-i+1)*i)
    call gemm (A1,X,Z)
    X=Z
    Y=Y+c*X
    D=D+sgn*c*X
    sgn=-sgn
  end do
!
  call gesv(a=D,b=Y,info=ier)
  if (ier/=0) then
    call warni('expm LAPACK95 gesv failed',ier)
    call set_error(FUNC_LAPACK_ER)
    deallocate(A1,D,Y,X,Z,work)
    return
  end if
!
  do i=1,j
    call gemm(Y,Y, Z)
    Y=Z
  end do
!
  B=Y
  deallocate(A1,D,Y,X,Z,work,stat=ier)
  if (ier>0) then
    call warn('expm dealloc failed')
    call set_error(FUNC_MEM_FREE_ER)
    return
  end if
!
  call set_error(FUNC_OK)
  return
end subroutine expm


end module func
