!
! Poroelastostatic Green's Tensor 
! for displacement response from a
! concentrated center of dilatation 
! within a reservoir
!
! Depth extrapolation method for
! a Horizontally-layered medium.
!
!
! Adjoint not implemented yet (04/19/2012)
!
! 2012 Musa Maharramov
! Stanford Exploration Project
!
module poroelastic_depth
use poroelastic_resmodel
use poroelastic_green
use error
use func
use simple_vector
use mkl95_blas
use mkl95_lapack
use mkl_dfti

implicit none
!
real, private, parameter :: stpi =  &
  16.*3.1415926535897932384626433832795028
real, private, parameter :: twopi =  &
  2.*3.1415926535897932384626433832795028
! accuracy of exp{A} evaluation
real, private, parameter :: EXP_ACCY = 1.e-12 
!
! poroelastic Green's tensor extends
! the base operator type
!
! note that the medium parameters and reservoir
! geometry are defined in the operator type
type, public, extends(pe_green) :: pe_depth
contains
  procedure :: apply => extrap
  procedure :: init => init_pe_depth
  procedure :: G => pedepthgreen
! medium parameter (see the paper)
! real, dimension(:,:,:), pointer :: c
! real, dimension(:,:,:,:,:,:,:), pointer :: greenmat
! are defined in the parent class
!
! horizontal wavenumbers precomputed for efficiency
  real, dimension(:), pointer :: kx, ky
! resolution parameters (required to avoid blow-up)
  integer :: maxikx, maxiky
end type  pe_depth

contains 

! compute 3D displacement d(1:3,maxz,x,y,rz,rx,ry)
! at (x,y,z) due to a concentrated dilatation
! at (rx,ry,rz) for the given reservoir model
! 
! Green's tensor is computed at the maximum
! model depth (usu above the reservoir)
! zmax=z0+(nz-1)*dz
  subroutine pedepthgreen(this,rz0,rdz,rnz,d)
  class(pe_depth), intent(inout) :: this
! z-grid where dilatation is applied
  real, intent(in) :: rdz
! allow reservoir z grid to have variable orig
  real, intent(inout), dimension(:,:) :: rz0
  integer, intent(in) :: rnz
! displacement(1:3,z,x,y,rz,rx,ry)
  real,intent(inout),dimension(:,:,:,:,:,:,:) :: d
! aux vars
  real :: x,y,z,rx,ry,rz,temp,bup
  real :: r1,r2
  integer :: ix,iy,rix,riy,riz,i
! temp displacement field
  real, dimension(:,:,:,:,:), allocatable :: u

 if (.not.associated(this%res)) return
  
associate (tres=>this%res)
select type(tres)
class is (pe_reservoir)

  associate(dx=>tres%dx,dy=>tres%dy,dz=>tres%dz, &
    rdx=>tres%rdx,rdy=>tres%rdy, &
    x0=>tres%x0,y0=>tres%y0,z0=>tres%z0, & 
    rx0=>tres%rx0,ry0=>tres%ry0, &
    nx=>tres%nx,ny=>tres%ny,nz=>tres%nz, &
    rnx=>tres%rnx,rny=>tres%rny)
!
! padding required for numeric differentiation
! of Mindlin's displacement tensor components wrt
! source coordinates later, hence the larger grid
  allocate(u(1:3,1:3,1:rnz+2,1:rnx+2,1:rny+2))
!
! all displacement points outside 
! of the reservoir
  DISP: do iy=1,ny
    y=dy*(iy-1)+y0
    do ix=1,nx
      x=dx*(ix-1)+x0
      z=dz*(nz-1)+z0
!
! all source points inside the reservoir
!$omp parallel do private(rix,riz,riy,rx,ry,rz,r1,r2,temp,bup)
      SRC: do riy=1,rny+2
        ry=rdy*(riy-2)+ry0
        do rix=1,rnx+2
          rx=rdx*(rix-2)+rx0
          do riz=1,rnz+2
            rz=rdz*(riz-2)+rz0(rix,riy)
            !
            r1=((x-rx)**2+(y-ry)**2+(z-rz)**2)**.5
            r2=((x-rx)**2+(y-ry)**2+(z+rz)**2)**.5
!
! blow-up control for the mirror-image
            bup=r2-z-rz
            bup=max(bup,min(rdx,rdy,dx,dy))
!
            temp= &
(z-rz)/(r1**3) + (3.-4.*tres%nu(nz,ix,iy))*(z-rz)/(r2**3) + &
4.*(1.-tres%nu(nz,ix,iy))*(1.-2.*tres%nu(nz,ix,iy))/(r2*bup)+ &
6.*z*rz*(z+rz)/(r2**5)
!

! x-displacement due to vertical force
u(3,1,riz,rix,riy)= (x-rx)*temp
! y-displacement due to vertical force
u(3,2,riz,rix,riy)= (y-ry)*temp
! z-displacement due to vertical force
u(3,3,riz,rix,riy)= &
(3.-4.*tres%nu(nz,ix,iy))/r1+ &
(5.-12.*tres%nu(nz,ix,iy)+8.*(tres%nu(nz,ix,iy)**2))/r2+&
((z-rz)**2)/(r1**3)+ &
((3.-4.*tres%nu(nz,ix,iy))*((z+rz)**2)-2.*z*rz)/(r2**3)+ &
6.*z*rz*((z+rz)**2)/(r2**5)
!
! x-displacement due to x-force
u(1,1,riz,rix,riy)= &
(3.-4.*tres%nu(nz,ix,iy))/r1+1./r2+((x-rx)**2)/(r1**3)+ &
(3.-4.*tres%nu(nz,ix,iy))*((x-rx)**2)/(r2**3)+ &
2.*(r2**2-3.*(x-rx)**2)*z*rz/(r2**5)+&
4.*(1.-tres%nu(nz,ix,iy))*(1.-2.*tres%nu(nz,ix,iy))* &
(r2**2-(x-rx)**2-r2*(z+rz))/(r2*(bup**2))
!
! y-displacement due to x-force
u(1,2,riz,rix,riy)= (x-rx)*(y-ry)*( &
1./(r1**3)+(3.-4.*tres%nu(nz,ix,iy))/(r2**3)-&
6.*z*rz/(r2**5)-&
4.*(1.-tres%nu(nz,ix,iy))*(1.-2.*tres%nu(nz,ix,iy))/&
(r2*(bup**2)) &
)
!
! z-displacement due to x-force
u(1,3,riz,rix,riy)= (x-rx)*( &
(z-rz)/(r1**3)+(3.-4.*tres%nu(nz,ix,iy))*(z-rz)/(r2**3)-&
4.*(1.-tres%nu(nz,ix,iy))*(1.-2.*tres%nu(nz,ix,iy))/&
(r2*bup)- 6.*z*rz*(z+rz)/(r2**5) &
)
! y-displacement due to y-force
u(2,2,riz,rix,riy)= &
(3.-4.*tres%nu(nz,ix,iy))/r1+1./r2+((y-ry)**2)/(r1**3)+ &
(3.-4.*tres%nu(nz,ix,iy))*((y-ry)**2)/(r2**3)+ &
2.*(r2**2-3.*(y-ry)**2)*z*rz/(r2**5)+&
4.*(1.-tres%nu(nz,ix,iy))*(1.-2.*tres%nu(nz,ix,iy))* &
(r2**2-(y-ry)**2-r2*(z+rz))/(r2*(bup**2))
!
! x-displacement due to y-force
u(2,1,riz,rix,riy)= (x-rx)*(y-ry)*( &
1./(r1**3)+(3.-4.*tres%nu(nz,ix,iy))/(r2**3)-&
6.*z*rz/(r2**5)-&
4.*(1.-tres%nu(nz,ix,iy))*(1.-2.*tres%nu(nz,ix,iy))/&
(r2*(bup**2)) &
)
!
! z-displacement due to y-force
u(2,3,riz,rix,riy)= (y-ry)*( &
(z-rz)/(r1**3)+(3.-4.*tres%nu(nz,ix,iy))*(z-rz)/(r2**3)-&
4.*(1.-tres%nu(nz,ix,iy))*(1.-2.*tres%nu(nz,ix,iy))/&
(r2*bup)- 6.*z*rz*(z+rz)/(r2**5) &
)
          end do
        end do
      end do SRC 
!$omp end parallel do
!     now compute the divergence
!$omp parallel do private(rix,riz,riy,i)
      DIVR: do riy=1,rny
        do rix=1,rnx
          do riz=1,rnz
            do i=1,3
! use central differences              
!             note for this class the stored Green's function
!             is just for a single (deepest) depth slice
              d(i,1,ix,iy,riz,rix,riy)= &
    (u(1,i,riz+1,rix+2,riy+1)-u(1,i,riz+1,rix,riy+1))/(2.*rdx)+ &
    (u(2,i,riz+1,rix+1,riy+2)-u(2,i,riz+1,rix+1,riy))/(2.*rdy)+ &
    (u(3,i,riz+2,rix+1,riy+1)-u(3,i,riz,rix+1,riy+1))/(2.*rdz)
            end do
          end do
        end do
      end do DIVR
!$omp end parallel do
    end do
  end do DISP
  deallocate(u)
end associate
!
class default
return
end select
end associate

end subroutine pedepthgreen
!
!
! Deformation modelling/
! displacement extrapolation operator
!
! d is displacement vector components
! packed in (x,y,z) triples
subroutine extrap(this, m, d, adj) 
implicit none
! the poroelastic green operator
  class(pe_depth), intent(inout) :: this
! pressure and displacement
  class(ls_vector), intent(inout) :: m, d
! adjoint flag
  logical, intent(in) :: adj
! aux vars
  real :: temp,tempx,tempy,tempz,x,y,z,rx,ry
  real :: oon,r2,rx2,ry2
  integer :: ix,iy,rix,riy,ind,rind,ier,iz,i,j,L(2)
!
! extrapolation matrices
complex, dimension (:,:,:,:), allocatable :: mat
complex, dimension (:,:,:), allocatable :: rhs
! 1D buffers for 2D FFT
complex, dimension(:), allocatable ::  &
  temp1, temp2, temp3
character(len=128) :: s
! FFT plan descriptor
type(DFTI_DESCRIPTOR), POINTER :: My_Desc_Handle

if (.not.associated(this%res) .or. &
    .not.associated(this%c)) return

!
! use the reservoir model this%res
! associated with the operator
associate(dx=>this%res%dx,dy=>this%res%dy, &
dz=>this%res%dz, &
rdx=>this%res%rdx,rdy=>this%res%rdy, &
x0=>this%res%x0,y0=>this%res%y0, &
z0=>this%res%z0, & 
rx0=>this%res%rx0,ry0=>this%res%ry0, &
nx=>this%res%nx,ny=>this%res%ny, &
nz=>this%res%nz, &
rnx=>this%res%rnx,rny=>this%res%rny, &
tres=>this%res)

select type(tres)
class is (pe_reservoir)

select type (m)
class is (vector)
  select type(d)
  class is (vector)
    select type(m)
    class is (vector)
!
      call d%zero();
!$omp parallel do private(x,y,z,ix,iy,rix,riy,ind,rind,temp, &
!$ tempx,tempy,tempz)
      do ix=1,nx
        x=dx*(ix-1)+x0
        do iy=1,ny
          y=dy*(iy-1)+y0
          z=dz*(nz-1)+z0
          ind=3*(nz*nx*(iy-1)+nz*(ix-1)+nz-1)+1
          do rix=1,rnx
            rx=rdx*(rix-1)+rx0
            do riy=1,rny
              ry=rdy*(riy-1)+ry0
! proceed only if internal point of the reservoir                  
              if (this%res%rzlo(rix,riy)> &
                this%res%rzhi(rix,riy)) cycle 
! index into the packed pore array
              rind=rnx*(riy-1)+rix
!
              temp=(this%res%rzhi(rix,riy)- &
          this%res%rzlo(rix,riy))*this%c(nz,ix,iy)
!
              if (this%mx) then
                tempx=this%greenmat(1,1,ix,iy,1,rix,riy)*temp
              else
                tempx=0.
              end if
              if (this%my) then
                tempy=this%greenmat(2,1,ix,iy,1,rix,riy)*temp
              else
                tempy=0.
              end if
              if (this%mz) then
                tempz=this%greenmat(3,1,ix,iy,1,rix,riy)*temp
              else
                tempz=0.
              end if
!
              d%val(ind)=d%val(ind)+tempx*m%val(rind)
              d%val(ind+1)=d%val(ind+1)+tempy*m%val(rind)
              d%val(ind+2)=d%val(ind+2)+tempz*m%val(rind)
            end do
          end do
        end do
      end do
!$omp end parallel do
      deallocate(this%greenmat)
!     now extrapolate the deformations upward from depth nz
      if (nz==1) return
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      allocate(rhs(6*nz,nx,ny),temp1(nx*ny),& 
     mat(6*nz,6*nz,nx,ny),temp2(nx*ny),temp3(nx*ny), stat=ier)
      if (ier/=0) then
        call warn('poroelastic_depth::extrap alloc(mat,rhs,temp*)')
        call set_error(PE_MEM_ALLOC_ER)
        return
      end if
!
!$omp parallel do private(ix,iy,ind)
      do ix=1,nx
        do iy=1,ny
          ind=3*(nz*nx*(iy-1)+nz*(ix-1)+nz-1)+1
          temp1(ix+(iy-1)*nx)=d%val(ind)
          temp2(ix+(iy-1)*nx)=d%val(ind+1)
          temp3(ix+(iy-1)*nx)=d%val(ind+2)
        end do  
      end do
!$omp end parallel do
!
!     FFT the data
      L(1)=nx; L(2)=ny
      ier = DftiCreateDescriptor( My_Desc_Handle, DFTI_SINGLE,&
          DFTI_COMPLEX, 2, L)
      ier = DftiCommitDescriptor( My_Desc_Handle)
      ier = DftiComputeForward( My_Desc_Handle, temp1)
      ier = DftiFreeDescriptor(My_Desc_Handle)
!
      ier = DftiCreateDescriptor( My_Desc_Handle, DFTI_SINGLE,&
          DFTI_COMPLEX, 2, L)
      ier = DftiCommitDescriptor( My_Desc_Handle)
      ier = DftiComputeForward( My_Desc_Handle, temp2)
      ier = DftiFreeDescriptor(My_Desc_Handle)
!
      ier = DftiCreateDescriptor( My_Desc_Handle, DFTI_SINGLE,&
          DFTI_COMPLEX, 2, L)
      ier = DftiCommitDescriptor( My_Desc_Handle)
      ier = DftiComputeForward( My_Desc_Handle, temp3)
      ier = DftiFreeDescriptor(My_Desc_Handle)
!
!$omp parallel do private(ix,iy,iz,i,j,temp,oon,rx2,ry2,r2,s)
      do ix=1,nx
        do iy=1,ny
          rhs(:,ix,iy)=0; mat(:,:,ix,iy)=0.
          rx2=this%kx(ix)**2;ry2=this%ky(iy)**2
          r2=rx2+ry2
!         construct a sparse matrix to solve BVP for each wn
          mat(:,:,ix,iy)=0; rhs(:,ix,iy)=0.
          rhs(6*(nz-1)+1,ix,iy)=temp1(ix+(iy-1)*nx)
          rhs(6*(nz-1)+2,ix,iy)=temp2(ix+(iy-1)*nx)
          rhs(6*(nz-1)+3,ix,iy)=temp3(ix+(iy-1)*nx)
!
          DOMAT: do iz=nz,1,-1
            oon=1./(1.-2.*tres%nu(iz,ix,iy))
            temp=1./(1.+oon)
!           v'_123=v_456
            if (iz>1 .and. iz<nz) then
              do j=1,3
                mat(6*(iz-1)+j,6*(iz)+j,ix,iy)=mat(6*(iz-1)+j,6*(iz)+j,ix,iy)+.5/dz
                mat(6*(iz-1)+j,6*(iz-2)+j,ix,iy)=mat(6*(iz-1)+j,6*(iz-2)+j,ix,iy)-.5/dz
                mat(6*(iz-1)+j,6*(iz-1)+j+3,ix,iy)=mat(6*(iz-1)+j,6*(iz-1)+j+3,ix,iy)-1.
              end do
!             just v'_456
              do j=4,6
                mat(6*(iz-1)+j,6*(iz)+j,ix,iy)=mat(6*(iz-1)+j,6*(iz)+j,ix,iy)+.5/dz
                mat(6*(iz-1)+j,6*(iz-2)+j,ix,iy)=mat(6*(iz-1)+j,6*(iz-2)+j,ix,iy)-.5/dz
              end do
            end if
!           rhs's for v'456
            mat(6*(iz-1)+4,6*(iz-1)+1,ix,iy)=mat(6*(iz-1)+4,6*(iz-1)+1,ix,iy)-(r2+rx2*oon)
            mat(6*(iz-1)+4,6*(iz-1)+2,ix,iy)=mat(6*(iz-1)+4,6*(iz-1)+2,ix,iy)-this%kx(ix)*this%ky(iy)*oon
            mat(6*(iz-1)+4,6*(iz-1)+6,ix,iy)=mat(6*(iz-1)+4,6*(iz-1)+6,ix,iy)+cmplx(0.,this%kx(ix)*oon)
!
            mat(6*(iz-1)+5,6*(iz-1)+1,ix,iy)=-this%kx(ix)*this%ky(iy)*oon
            mat(6*(iz-1)+5,6*(iz-1)+2,ix,iy)=-(r2+ry2*oon)
            mat(6*(iz-1)+5,6*(iz-1)+6,ix,iy)=cmplx(0.,this%ky(iy)*oon)
!
            mat(6*(iz-1)+6,6*(iz-1)+3,ix,iy)= &
   mat(6*(iz-1)+6,6*(iz-1)+3,ix,iy)-r2*temp
            mat(6*(iz-1)+6,6*(iz-1)+4,ix,iy)= & 
   mat(6*(iz-1)+6,6*(iz-1)+4,ix,iy)+cmplx(0.,this%kx(ix)*oon*temp)
            mat(6*(iz-1)+6,6*(iz-1)+5,ix,iy)= &
   mat(6*(iz-1)+6,6*(iz-1)+5,ix,iy)+cmplx(0.,this%ky(iy)*oon*temp)
          end do DOMAT
!
!         first and last 3 equations
!         v'_123=v_456 at zmin
          do j=1,3
!           2nd order FD
            mat(j,j,ix,iy)=mat(j,j,ix,iy)-1.5/dz
            mat(j,6+j,ix,iy)=mat(j,6+j,ix,iy)+2./dz
            mat(j,12+j,ix,iy)=mat(j,12+j,ix,iy)-.5/dz
!           1st-order FD
!            mat(j,6+j,ix,iy)=mat(j,6+j,ix,iy)+1./dz
!            mat(j,j,ix,iy)=mat(j,j,ix,iy)-1./dz
            mat(j,j+3,ix,iy)=mat(j,j+3,ix,iy)-1.
          end do
!         just v'_456 at zmax
          do j=4,6
!           2nd-order FD
            mat(6*(nz-1)+j,6*(nz-1)+j,ix,iy)=mat(6*(nz-1)+j,6*(nz-1)+j,ix,iy)+1.5/dz
            mat(6*(nz-1)+j,6*(nz-2)+j,ix,iy)=mat(6*(nz-1)+j,6*(nz-2)+j,ix,iy)-2./dz
            mat(6*(nz-1)+j,6*(nz-3)+j,ix,iy)=mat(6*(nz-1)+j,6*(nz-3)+j,ix,iy)+.5/dz
!           1st-order FD
!            mat(6*(nz-1)+j,6*(nz-1)+j,ix,iy)=mat(6*(nz-1)+j,6*(nz-1)+j,ix,iy)+1./dz
!            mat(6*(nz-1)+j,6*(nz-2)+j,ix,iy)=mat(6*(nz-1)+j,6*(nz-2)+j,ix,iy)-1./dz
          end do

!         initial conditions
          mat(6*nz-5,6*nz-5,ix,iy)=1.;mat(6*nz-4,6*nz-4,ix,iy)=1.
          mat(6*nz-3,6*nz-3,ix,iy)=1.
!
!         free boundary conditions
          temp=2*tres%nu(1,ix,iy)/(1.+2.*tres%nu(1,ix,iy))
          mat(4,4,ix,iy)=1.; mat(5,5,ix,iy)=1.; mat(6,6,ix,iy)=1.
          mat(4,3,ix,iy)=cmplx(0.,this%kx(ix))
          mat(5,3,ix,iy)=cmplx(0.,this%ky(iy))
          mat(6,1,ix,iy)=cmplx(0.,this%kx(ix)*temp/(2.+temp))
          mat(6,2,ix,iy)=cmplx(0.,this%ky(iy)*temp/(2.+temp))
!
!         inter-layer "stitching"
          do iz=nz,1,-1
            do i=1,6
              do j=1,6
                mat(6*(iz-1)+i,6*(iz-1)+j,ix,iy)= &
     tres%mu(iz,ix,iy)*mat(6*(iz-1)+i,6*(iz-1)+j,ix,iy)
                if (iz>1) then
!                 consider shear modulus variable
!                 only when differentiating the stresses
                  if (i>3) then
                    mat(6*(iz-1)+i,6*(iz-2)+j,ix,iy)= &
     tres%mu(iz-1,ix,iy)*mat(6*(iz-1)+i,6*(iz-2)+j,ix,iy)
                  else
                    mat(6*(iz-1)+i,6*(iz-2)+j,ix,iy)= &
     tres%mu(iz,ix,iy)*mat(6*(iz-1)+i,6*(iz-2)+j,ix,iy)
                  end if
                end if
                if (iz<nz) then
!                 consider shear modulus variable
!                 only when differentiating the stresses
                  if (i>3) then
                    mat(6*(iz-1)+i,6*iz+j,ix,iy)= &
     tres%mu(iz+1,ix,iy)*mat(6*(iz-1)+i,6*iz+j,ix,iy)
                  else
                    mat(6*(iz-1)+i,6*iz+j,ix,iy)= &
     tres%mu(iz,ix,iy)*mat(6*(iz-1)+i,6*iz+j,ix,iy)
                  end if
                end if
              end do
              rhs(6*(iz-1)+i,ix,iy)=rhs(6*(iz-1)+i,ix,iy)*tres%mu(iz,ix,iy)
            end do
          end do
!
!         solve the sparse system
          call gesv(a=mat(:,:,ix,iy),b=rhs(:,ix,iy),info=j)
          if (j/=0) then
!$omp critical
            call warni('extrap LAPACK95 gesv failed',j)
!$omp end critical
          end if
!
        end do
      end do
!
!     inverse-FFT and copy to 'd'
      COPYDEPTHS: do iz=1,nz
!$omp parallel do private(ix,iy)
        do ix=1,nx
          do iy=1,ny
!           crude filter
            if (ix>this%maxikx .and. ix<nx-this%maxikx .or.  &
                iy>this%maxiky .and. iy<ny-this%maxiky) then
!              rhs(6*iz-5:6*iz,ix,iy)=0.
            end if
            temp1(ix+(iy-1)*nx)=rhs(6*(iz-1)+1,ix,iy)
            temp2(ix+(iy-1)*nx)=rhs(6*(iz-1)+2,ix,iy)
            temp3(ix+(iy-1)*nx)=rhs(6*(iz-1)+3,ix,iy)
          end do  
        end do
!$omp end parallel do
! 
        L(1)=nx; L(2)=ny
        temp=1./real(nx*ny)
        ier = DftiCreateDescriptor( My_Desc_Handle, DFTI_SINGLE,&
          DFTI_COMPLEX, 2, L)
        ier = DftiSetValue(My_Desc_Handle, DFTI_BACKWARD_SCALE,temp) 
        ier = DftiCommitDescriptor( My_Desc_Handle)
        ier = DftiComputeBackward( My_Desc_Handle, temp1)
        ier = DftiFreeDescriptor(My_Desc_Handle)
!
        ier = DftiCreateDescriptor( My_Desc_Handle, DFTI_SINGLE,&
          DFTI_COMPLEX, 2, L)
        ier = DftiSetValue(My_Desc_Handle, DFTI_BACKWARD_SCALE,temp) 
        ier = DftiCommitDescriptor( My_Desc_Handle)
        ier = DftiComputeBackward( My_Desc_Handle, temp2)
        ier = DftiFreeDescriptor(My_Desc_Handle)
!
        ier = DftiCreateDescriptor( My_Desc_Handle, DFTI_SINGLE,&
          DFTI_COMPLEX, 2, L)
        ier = DftiSetValue(My_Desc_Handle, DFTI_BACKWARD_SCALE,temp) 
        ier = DftiCommitDescriptor( My_Desc_Handle)        
        ier = DftiComputeBackward( My_Desc_Handle, temp3)
        ier = DftiFreeDescriptor(My_Desc_Handle)
!
!     copy the result back to the displacement field
!$omp parallel do private(ix,iy,ind)
        do ix=1,nx
          do iy=1,ny
            ind=3*(nz*nx*(iy-1)+nz*(ix-1)+iz-1)+1
            d%val(ind)=temp1(ix+(iy-1)*nx)
            d%val(ind+1)=temp2(ix+(iy-1)*nx)
            d%val(ind+2)=temp3(ix+(iy-1)*nx)
          end do  
        end do
!$omp end parallel do
!
      end do COPYDEPTHS

      deallocate(mat,rhs,temp1,temp2,temp3,stat=ier)
      if (ier/=0) then
        call warn('poroelastic_depth::extrap dealloc(mat,rhs,temp*)')
        call set_error(PE_MEM_FREE_ER)
        return
      end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    class default
    end select
  class default
  return
  end select
end select
class default
return
end select
end associate
return

end subroutine extrap
!
!
! initialise poroelastostatic
! modelling operator
subroutine init_pe_depth(this,res,mx,my,mz)
implicit none
class(pe_depth), intent(inout) :: this
class(base_reservoir),intent(inout),pointer :: res
logical, intent(in) :: mx, my, mz
! aux vars
integer :: ix,iy,iz
! reservoir "middle surface"
real, dimension(:,:),allocatable :: rz0
real :: temp

if (.not.associated(res)) return
!
! associate reservoir 'res' with
! this operator
select type(res)
class is (pe_reservoir)
  this%res => res
  this%mx=mx;this%my=my;this%mz=mz

  associate(dx=>res%dx,dy=>res%dy,dz=>res%dz, &
    rdx=>res%rdx,rdy=>res%rdy, &
    x0=>res%x0,y0=>res%y0,z0=>res%z0, & 
    rx0=>res%rx0,ry0=>res%ry0, &
    nx=>res%nx,ny=>res%ny,nz=>res%nz, &
    rnx=>res%rnx,rny=>res%rny)
    if (associated(this%c)) deallocate(this%c)
    allocate(this%c(1:nz,1:nx,1:ny))
    if (associated(this%greenmat)) &
      deallocate(this%greenmat)
!   note for 'poroelastic_depth' greenmat
!   is defined for only one (deepest) 
!   horizontal slice
    allocate(this%greenmat(1:3,1:1,& 
       1:nx,1:ny,1:1,1:rnx,1:rny))
    allocate(rz0(rnx+2,rny+2))
!   horiz wn's
    allocate(this%kx(nx),this%ky(ny))
!
    do ix=1,this%res%nx/2
      this%kx(ix)=twopi*(ix-1)/(res%dx*res%nx)
    end do
    do ix=this%res%nx/2+1,this%res%nx
      this%kx(ix)=twopi*(ix-1-this%res%nx)/(res%dx*res%nx)
    end do
!
    do iy=1,this%res%ny/2
      this%ky(iy)=twopi*(iy-1)/(res%dy*res%ny)
    end do
    do iy=this%res%ny/2+1,this%res%ny
      this%kx(iy)=twopi*(iy-1-this%res%ny)/(res%dy*res%ny)
    end do
!   auto set set spatial resolution as that 
!   of the reservoir grid 
    this%maxikx=max(5,this%res%rnx/4)
    this%maxiky=max(5,this%res%rny/4)

!$omp parallel do private(ix,iy,iz,temp)
    do iy=1,ny
      do ix=1,nx
        do iz=1,nz
!  Biot coefficient over 16 pi       
          temp= &
   (1.-res%Ksat(iz,ix,iy)/res%Kmin(iz,ix,iy))/stpi
          this%c(iz,ix,iy)=temp*rdx*rdy/ &
   (res%mu(iz,ix,iy)*(1.-res%nu(iz,ix,iy)))
        end do
      end do
    end do
!$omp end parallel do
    do iy=1,rny
      do ix=1,rnx
        rz0(ix+1,iy+1)=(res%rzlo(ix,iy)+res%rzhi(ix,iy))/2.
      end do
    end do
    rz0(1,1:rny+2)=rz0(2,rny/2)
    rz0(rnx+2,1:rny+2)=rz0(rnx+1,rny/2)
    rz0(1:rnx+2,1)=rz0(rnx/2,2)
    rz0(1:rnx+2,rny+2)=rz0(rnx/2,rny+1)
!
    write (0,*)'COMPUTING GREEN'
    call this%G(rz0,min(rdx,rdy),1,this%greenmat)
    deallocate(rz0)
  end associate
class default
!
end select
return
end subroutine init_pe_depth

end module poroelastic_depth


