!
! Poroelastostatic Green's Tensor 
! for displacement response from a
! concentrated center of dilatation 
! within a reservoir
!
! 2012 Musa Maharramov
! Stanford Exploration Project
!
module poroelastic_green
use poroelastic_resmodel
use green_tensor
use simple_vector
implicit none
!
real, private, parameter :: stpi =  &
  16.*3.1415926535897932384626433832795028

! poroelastic Green's tensor extends
! the base operator type
!
! note that the medium parameters and reservoir
! geometry are defined in the operator type
type, public, extends(green) :: pe_green
contains
  procedure :: apply => deform
  procedure :: init => init_pe_green
  procedure :: G => pegreen

! medium parameter (see the paper)
real, dimension(:,:,:), pointer :: c
real, dimension(:,:,:,:,:,:,:), pointer :: greenmat
end type  pe_green

contains 


! compute 3D displacement d(1:3,z,x,y,rz,rx,ry)
! at (x,y,z) due to a concentrated dilatation
! at (rx,ry,rz) for the given reservoir model
  subroutine pegreen(this,rz0,rdz,rnz,d)
  class(pe_green), intent(inout) :: this
! z-grid where dilatation is applied
  real, intent(in) :: rdz
! allow reservoir z grid to have variable orig
  real, intent(inout), dimension(:,:) :: rz0
  integer, intent(in) :: rnz
! displacement(1:3,z,x,y,rz,rx,ry)
  real,intent(inout),dimension(:,:,:,:,:,:,:) :: d
! aux vars
  real :: x,y,z,rx,ry,rz,temp,bup
  real :: r1,r2
  integer :: ix,iy,iz,rix,riy,riz,i
! temp displacement field
  real, dimension(:,:,:,:,:), allocatable :: u

if (.not.associated(this%res)) return
  
associate (tres=>this%res)
select type(tres)
class is (pe_reservoir)

  associate(dx=>tres%dx,dy=>tres%dy,dz=>tres%dz, &
    rdx=>tres%rdx,rdy=>tres%rdy, &
    x0=>tres%x0,y0=>tres%y0,z0=>tres%z0, & 
    rx0=>tres%rx0,ry0=>tres%ry0, &
    nx=>tres%nx,ny=>tres%ny,nz=>tres%nz, &
    rnx=>tres%rnx,rny=>tres%rny)
!
! padding required for numeric differentiation
! of Mindlin's displacement tensor components wrt
! source coordinates later, hence the larger grid
  allocate(u(1:3,1:3,1:rnz+2,1:rnx+2,1:rny+2))
!
! all displacement points outside 
! of the reservoir
  DISP: do iy=1,ny
    y=dy*(iy-1)+y0
    do ix=1,nx
      x=dx*(ix-1)+x0
      do iz =1,nz
        z=dz*(iz-1)+z0
!
! all source points inside the reservoir
!$omp parallel do private(rix,riz,riy,rx,ry,rz,r1,r2,temp,bup)
        SRC: do riy=1,rny+2
          ry=rdy*(riy-2)+ry0
          do rix=1,rnx+2
            rx=rdx*(rix-2)+rx0
            do riz=1,rnz+2
              rz=rdz*(riz-2)+rz0(rix,riy)
              !
              r1=((x-rx)**2+(y-ry)**2+(z-rz)**2)**.5
              r2=((x-rx)**2+(y-ry)**2+(z+rz)**2)**.5
!
! blow-up control for the mirror-image
              bup=r2-z-rz
              bup=max(bup,min(dx,dy,rdx,rdy))
!
              temp= &
(z-rz)/(r1**3) + (3.-4.*tres%nu(iz,ix,iy))*(z-rz)/(r2**3) + &
4.*(1.-tres%nu(iz,ix,iy))*(1.-2.*tres%nu(iz,ix,iy))/(r2*bup)+ &
6.*z*rz*(z+rz)/(r2**5)
!

! x-displacement due to vertical force
u(3,1,riz,rix,riy)= (x-rx)*temp
! y-displacement due to vertical force
u(3,2,riz,rix,riy)= (y-ry)*temp
! z-displacement due to vertical force
u(3,3,riz,rix,riy)= &
(3.-4.*tres%nu(iz,ix,iy))/r1+ &
(5.-12.*tres%nu(iz,ix,iy)+8.*(tres%nu(iz,ix,iy)**2))/r2+&
((z-rz)**2)/(r1**3)+ &
((3.-4.*tres%nu(iz,ix,iy))*((z+rz)**2)-2.*z*rz)/(r2**3)+ &
6.*z*rz*((z+rz)**2)/(r2**5)
!
! x-displacement due to x-force
u(1,1,riz,rix,riy)= &
(3.-4.*tres%nu(iz,ix,iy))/r1+1./r2+((x-rx)**2)/(r1**3)+ &
(3.-4.*tres%nu(iz,ix,iy))*((x-rx)**2)/(r2**3)+ &
2.*(r2**2-3.*(x-rx)**2)*z*rz/(r2**5)+&
4.*(1.-tres%nu(iz,ix,iy))*(1.-2.*tres%nu(iz,ix,iy))* &
(r2**2-(x-rx)**2-r2*(z+rz))/(r2*(bup**2))
!
! y-displacement due to x-force
u(1,2,riz,rix,riy)= (x-rx)*(y-ry)*( &
1./(r1**3)+(3.-4.*tres%nu(iz,ix,iy))/(r2**3)-&
6.*z*rz/(r2**5)-&
4.*(1.-tres%nu(iz,ix,iy))*(1.-2.*tres%nu(iz,ix,iy))/&
(r2*(bup**2)) &
)
!
! z-displacement due to x-force
u(1,3,riz,rix,riy)= (x-rx)*( &
(z-rz)/(r1**3)+(3.-4.*tres%nu(iz,ix,iy))*(z-rz)/(r2**3)-&
4.*(1.-tres%nu(iz,ix,iy))*(1.-2.*tres%nu(iz,ix,iy))/&
(r2*bup)- 6.*z*rz*(z+rz)/(r2**5) &
)
! y-displacement due to y-force
u(2,2,riz,rix,riy)= &
(3.-4.*tres%nu(iz,ix,iy))/r1+1./r2+((y-ry)**2)/(r1**3)+ &
(3.-4.*tres%nu(iz,ix,iy))*((y-ry)**2)/(r2**3)+ &
2.*(r2**2-3.*(y-ry)**2)*z*rz/(r2**5)+&
4.*(1.-tres%nu(iz,ix,iy))*(1.-2.*tres%nu(iz,ix,iy))* &
(r2**2-(y-ry)**2-r2*(z+rz))/(r2*(bup**2))
!
! x-displacement due to y-force
u(2,1,riz,rix,riy)= (x-rx)*(y-ry)*( &
1./(r1**3)+(3.-4.*tres%nu(iz,ix,iy))/(r2**3)-&
6.*z*rz/(r2**5)-&
4.*(1.-tres%nu(iz,ix,iy))*(1.-2.*tres%nu(iz,ix,iy))/&
(r2*(bup**2)) &
)
!
! z-displacement due to y-force
u(2,3,riz,rix,riy)= (y-ry)*( &
(z-rz)/(r1**3)+(3.-4.*tres%nu(iz,ix,iy))*(z-rz)/(r2**3)-&
4.*(1.-tres%nu(iz,ix,iy))*(1.-2.*tres%nu(iz,ix,iy))/&
(r2*bup)- 6.*z*rz*(z+rz)/(r2**5) &
)
            end do
          end do
        end do SRC 
!$omp end parallel do
!       now compute the divergence
!$omp parallel do private(rix,riz,riy,i)
        DIVR: do riy=1,rny
          do rix=1,rnx
            do riz=1,rnz
              do i=1,3
! use central differences              
                d(i,iz,ix,iy,riz,rix,riy)= &
    (u(1,i,riz+1,rix+2,riy+1)-u(1,i,riz+1,rix,riy+1))/(2.*rdx)+ &
    (u(2,i,riz+1,rix+1,riy+2)-u(2,i,riz+1,rix+1,riy))/(2.*rdy)+ &
    (u(3,i,riz+2,rix+1,riy+1)-u(3,i,riz,rix+1,riy+1))/(2.*rdz)
              end do
            end do
          end do
        end do DIVR
!$omp end parallel do
      end do
    end do
  end do DISP
  deallocate(u)
end associate
!
class default
return
end select
end associate
end subroutine pegreen


! Deformation operator
! d is displacement vector components
! packed in (x,y,z) triples
subroutine deform(this, m, d, adj) 
implicit none
! the poroelastic green operator
class(pe_green), intent(inout) :: this
! pressure and displacement
class(ls_vector), intent(inout) :: m, d
! adjoint flag
logical, intent(in) :: adj
! aux vars
real :: temp,tempx,tempy,tempz,x,y,z,rx,ry
integer :: ix,iy,iz,rix,riy,ind,rind

if (.not.associated(this%res) .or. &
    .not.associated(this%c)) return
!
! use the reservoir model this%res
! associated with the operator
associate(dx=>this%res%dx,dy=>this%res%dy, &
dz=>this%res%dz, &
rdx=>this%res%rdx,rdy=>this%res%rdy, &
x0=>this%res%x0,y0=>this%res%y0, &
z0=>this%res%z0, & 
rx0=>this%res%rx0,ry0=>this%res%ry0, &
nx=>this%res%nx,ny=>this%res%ny, &
nz=>this%res%nz, &
rnx=>this%res%rnx,rny=>this%res%rny)

select type (m)
class is (vector)
  select type(d)
  class is (vector)
    select type(m)
    class is (vector)
!
!     compute the adjoint
      if (adj) then
        call m%zero()             
!$omp parallel do private(x,y,z,ix,iy,iz,rix,riy,ind, &
!$ rind,temp,tempx,tempy,tempz)
        do rix=1,rnx
          rx=rdx*(rix-1)+rx0
          do riy=1,rny
            ry=rdy*(riy-1)+ry0
! index into the packed pore array
            rind=rnx*(riy-1)+rix
! proceed only if internal point of the reservoir                             
            if (this%res%rzlo(rix,riy)> &
              this%res%rzhi(rix,riy)) cycle 
            do ix=1,nx
              x=dx*(ix-1)+x0
              do iy=1,ny
                y=dy*(iy-1)+y0
                do iz=1,nz
                  z=dz*(iz-1)+z0
! index into the displacement array
                  ind=3*(nz*nx*(iy-1)+nz*(ix-1)+iz-1)+1
!
                  temp=(this%res%rzhi(rix,riy)- &
                    this%res%rzlo(rix,riy))*this%c(iz,ix,iy)
!
                  if (this%mx) then
                    tempx=this%greenmat(1,iz,ix,iy,1,rix,riy)*temp
                  else
                    tempx=0.
                  end if
                  if (this%my) then
                    tempy=this%greenmat(2,iz,ix,iy,1,rix,riy)*temp
                  else
                    tempy=0.
                  end if
                  if (this%mz) then
                    tempz=this%greenmat(3,iz,ix,iy,1,rix,riy)*temp
                  else
                    tempz=0.
                  end if
!
                  m%val(rind)=m%val(rind)+tempx*d%val(ind)+ &
        tempy*d%val(ind+1)+tempz*d%val(ind+2)
                end do
              end do
            end do
          end do
        end do
!$omp end parallel do
      else              
        call d%zero();
!$omp parallel do private(x,y,z,ix,iy,iz,rix,riy,ind,rind,temp, &
!$ tempx,tempy,tempz)
        do ix=1,nx
          x=dx*(ix-1)+x0
          do iy=1,ny
            y=dy*(iy-1)+y0
            do iz=1,nz
              z=dz*(iz-1)+z0
              ind=3*(nz*nx*(iy-1)+nz*(ix-1)+iz-1)+1
              do rix=1,rnx
                rx=rdx*(rix-1)+rx0
                do riy=1,rny
                  ry=rdy*(riy-1)+ry0
! proceed only if internal point of the reservoir                  
                  if (this%res%rzlo(rix,riy)> &
                    this%res%rzhi(rix,riy)) cycle 
! index into the packed pore array
                  rind=rnx*(riy-1)+rix
!
                  temp=(this%res%rzhi(rix,riy)- &
          this%res%rzlo(rix,riy))*this%c(iz,ix,iy)
!
                  if (this%mx) then
                    tempx=this%greenmat(1,iz,ix,iy,1,rix,riy)*temp
                  else
                    tempx=0.
                  end if
                  if (this%my) then
                    tempy=this%greenmat(2,iz,ix,iy,1,rix,riy)*temp
                  else
                    tempy=0.
                  end if
                  if (this%mz) then
                    tempz=this%greenmat(3,iz,ix,iy,1,rix,riy)*temp
                  else
                    tempz=0.
                  end if
!
                  d%val(ind)=d%val(ind)+tempx*m%val(rind)
                  d%val(ind+1)=d%val(ind+1)+tempy*m%val(rind)
                  d%val(ind+2)=d%val(ind+2)+tempz*m%val(rind)
                end do
              end do
            end do
          end do
        end do
!$omp end parallel do
      end if
    class default
    end select
  class default
  return
  end select
end select
end associate
return
end subroutine deform

!
! initialise poroelastostatic
! operator
subroutine init_pe_green(this,res,mx,my,mz)
implicit none
class(pe_green), intent(inout) :: this
class(base_reservoir),intent(inout),pointer :: res
logical, intent(in) :: mx, my, mz
! aux vars
integer :: ix,iy,iz
! reservoir "middle surface"
real, dimension(:,:),allocatable :: rz0
real :: temp

if (.not.associated(res)) return
!
! associate reservoir 'res' with
! this operator
select type(res)
class is (pe_reservoir)
  this%res => res
  this%mx=mx;this%my=my;this%mz=mz

  associate(dx=>res%dx,dy=>res%dy,dz=>res%dz, &
    rdx=>res%rdx,rdy=>res%rdy, &
    x0=>res%x0,y0=>res%y0,z0=>res%z0, & 
    rx0=>res%rx0,ry0=>res%ry0, &
    nx=>res%nx,ny=>res%ny,nz=>res%nz, &
    rnx=>res%rnx,rny=>res%rny)
    if (associated(this%c)) deallocate(this%c)
    allocate(this%c(1:nz,1:nx,1:ny))
    if (associated(this%greenmat)) &
      deallocate(this%greenmat)
    allocate(this%greenmat(1:3,1:nz,& 
       1:nx,1:ny,1:1,1:rnx,1:rny))
    allocate(rz0(rnx+2,rny+2))
!
!$omp parallel do private(ix,iy,iz,temp)
    do iy=1,ny
      do ix=1,nx
        do iz=1,nz
!  Biot coefficient over 16 pi       
          temp= &
   (1.-res%Ksat(iz,ix,iy)/res%Kmin(iz,ix,iy))/stpi
          this%c(iz,ix,iy)=temp*rdx*rdy/ &
   (res%mu(iz,ix,iy)*(1.-res%nu(iz,ix,iy)))
        end do
      end do
    end do
!$omp end parallel do
    do iy=1,rny
      do ix=1,rnx
        rz0(ix+1,iy+1)=(res%rzlo(ix,iy)+res%rzhi(ix,iy))/2.
      end do
    end do
    rz0(1,1:rny+2)=rz0(2,rny/2)
    rz0(rnx+2,1:rny+2)=rz0(rnx+1,rny/2)
    rz0(1:rnx+2,1)=rz0(rnx/2,2)
    rz0(1:rnx+2,rny+2)=rz0(rnx/2,rny+1)
!
    write (0,*)'COMPUTING GREEN'
    call this%G(rz0,min(rdx,rdy),1,this%greenmat)
    deallocate(rz0)
  end associate
class default
!
end select
return
!
end subroutine init_pe_green

end module poroelastic_green


