!
! Generic Least Squares Solver abstract interface library
! for solving overdetermined Linear Least Squares problem:
!
!      ||Am-d||-> min                     (*)
!
! where m is an N-dimensional vector vector, d is an M-dimensional
! data vector and A is an M x N dimensional real matrix
! of maximum rank.
!
! 2012 Stanford University 
! Musa Maharramov, maharram@stanford.edu
!
module base_solver 
use base_vector
use base_oper
implicit none


! abstract solver class
type, abstract :: ls_solver
! polymorphic pointers to model and data vectors
  class(ls_vector), pointer, public :: m,d
! the residual
  class(ls_vector), pointer, public :: r 

! had to make the following two members
! public to allow access from derived class
! implementation
!
! pointer to the operator procedure
  class(ls_oper), pointer, public :: opr00
! iteration count
  integer, public :: iter = 0

!  type-bound procedures
contains
  procedure(init_), deferred, pass(this) :: init
  procedure(step_), deferred, pass(this) :: step
  procedure(getiter_),public, deferred, pass(this) :: get_iter
  procedure(stalled_),public, deferred, pass(this) :: stalled
end type ls_solver

!
abstract interface
! procedure interface for solver initialisation
  subroutine init_(this, op, m, d, resume, w, tol)
    import :: ls_vector, ls_solver, ls_oper
!
!   containing object
    class(ls_solver), intent(inout) :: this

!   pointer to a procedure implementing A and A*
    class(ls_oper), intent(in), target :: op

!   initial model (m) and data d (right-hand side of (*))
!   have to be inout for error logging
    class(ls_vector), intent(inout) :: m, d

!   resume flag - don't clear the state if .true.
    logical, optional :: resume
!
!   optional LS weights
    class(ls_vector), optional, intent(inout) :: w
!
!   tolerance parameter -- useful for multi-grid
    real, optional, intent(in) :: tol


  end subroutine init_

! procedure interface for solver stepper 
  subroutine step_(this)
  import ls_solver
!   containing object
    class(ls_solver), intent(inout) :: this
  end subroutine step_
!
  function getiter_(this)
  import :: ls_solver
  integer :: getiter_
  class(ls_solver), intent(in) :: this
  end function getiter_
!
!
  function stalled_(this)
  import :: ls_solver
  logical :: stalled_
  class(ls_solver), intent(in) :: this
  end function stalled_
end interface
 
end module base_solver
