! Sample solver implementatios for solving
! overdetermined Linear Least Squares problem:
!
!      ||Am-d||-> min                     (*)
!
! where m is an N-dimensional vector vector, d is an M-dimensional
! data vector and A is an M x N dimensional real matrix
! of maximum rank.
!
! 2012 Stanford Exploration Project
! Musa Maharramov, musa@sep.stanford.edu
!
module cgnr
use error
use base_vector
use base_solver
implicit none
!
! IMPLEMENTATION
!
! tolerance/round-up
real, public, parameter ::  DEFAULT_LS_TOL = 1.e-6
!
!
! a sample  solver implementation of the Conjugate
! Residuals on Normal Equations
!
! IMPORTANT: In a production solver, greater care 
! should be taken of the potential loss of
! orthogonality of Krylov-space bases.
!
! This class is intended just as a sample 
! demonstration of how to extend ls_solver
!
type, public, extends(ls_solver)  :: ls_cgnr
! auxiliary CR structures:
  class(ls_vector),private,pointer :: p,z,w,wt
!
! flag indicating a breakdown of Krylov iterations
  logical :: breakdown = .false.
! tolerance parameter
  real :: ls_tol
! weight flag
  logical :: hasw
contains
        procedure, public :: init => cgnr_init
        procedure, public :: step => cgnr_step
        procedure, public :: get_iter => cgnr_getiter
        procedure, public :: stalled => cgnr_stalled
end type ls_cgnr
!
contains
!
!
! check if the iterations have stalled
  function cgnr_stalled(this)
  logical :: cgnr_stalled
  class(ls_cgnr), intent(in) :: this
    cgnr_stalled=this%breakdown
    return
  end function cgnr_stalled
!
! return iteration count
  function cgnr_getiter(this)
  integer :: cgnr_getiter
  class(ls_cgnr), intent(in) :: this
    cgnr_getiter=this%iter
    return
  end function cgnr_getiter

! ls_cgnr solver initialisation
  subroutine cgnr_init(this, op, m, d, resume, w, tol)
!
!   containing object
    class(ls_cgnr), intent(inout) :: this

!   pointer to a procedure implementing A and A*
    class(ls_oper), intent(in), target :: op

!   initial model (m) and data d (right-hand side of (*))
!   have to be inout for error logging
    class(ls_vector), intent(inout) :: m, d
!
!   optional LS weights
    class(ls_vector), optional, intent(inout) :: w

!   resume flag - don't clear the state if .true.
    logical, optional :: resume
!
!   tolerance parameter -- useful for multi-grid
    real, optional, intent(in) :: tol

    if (.not.present(resume) .or. .not.resume) then
      call set_error(LS_OK); this%iter=0;
    end if
!
    if (present(tol)) then
      this%ls_tol=tol
    else
      this%ls_tol=DEFAULT_LS_TOL
    end if
!
    if (present(w)) then
      this%hasw=.true.
      call w%spawn(this%wt)
    else
      this%hasw=.false.
    end if

!   pointer to operator
    this%opr00=>op
!
!   init approx
    call m%spawn(this%m)
!   RHS
    call d%spawn(this%d)
!   apply weights
    if (this%hasw) then
      call this%d%cwmult(this%d,this%wt)
    end if
!
!   create the residual vector - will form new (r,m)=x
    call d%spawn(this%r);
!   auxiliary vectors p, z, w
    call m%spawn(this%p); call m%spawn(this%z) 
    call d%spawn(this%w)
!   r=d-Fm
    call this%opr00%apply(this%m,this%r,.false.)
!   apply weights
    if (this%hasw) then
      call this%r%cwmult(this%r,this%wt)
    end if
    
    call this%r%lc(this%d,this%r,1.,-1.)
!   z=FT r
!   apply weights
    if (this%hasw) then
      call this%r%cwmult(this%r,this%wt)
    end if
    call this%opr00%apply(this%z,this%r,.true.)
    this%p=this%z
!
    this%breakdown = .false.
!    
    return
!
  end subroutine cgnr_init

! ls_cgnr solver step 
  subroutine cgnr_step(this)
!   containing object
    class(ls_cgnr), intent(inout) :: this
!   aux vars
    real :: alpha, beta, temp

    this%iter=this%iter+1

!   w=Ap
    call this%opr00%apply(this%p,this%w,.false.)
!   apply weights
    if (this%hasw) then
      call this%w%cwmult(this%w,this%wt)
    end if
 
!   |w|^2
    temp=this%w%dot(this%w)
    if (abs(temp)<this%ls_tol) then
      this%breakdown=.true.
      return
    end if
!
    alpha=this%z%dot(this%z)/temp
!   r=r-alpha * w
    call this%r%lc(this%r,this%w,1.,-alpha)
!   m=m+alpha * p
    call this%m%lc(this%m,this%p,1.,alpha)
!   |z_i|^2
    temp=this%z%dot(this%z)
    if (abs(temp)<this%ls_tol) then
      this%breakdown=.true.
      return
    end if
!
!   z=FT r
!   apply weights
    if (this%hasw) then
      call this%r%cwmult(this%r,this%wt)
    end if
    call this%opr00%apply(this%z,this%r,.true.)
!
!   beta=|z_{i+1}|^2/|z_i|^2
    beta=this%z%dot(this%z)/temp
!   p=z+beta * p
    call this%p%lc(this%z,this%p,1.,beta)
!
    return
  end subroutine cgnr_step
!
end module cgnr
