! Sample solver implementatios for solving
! overdetermined Linear Least Squares problem:
!
!      ||Am-d||-> min                     (*)
!
! where m is an N-dimensional vector vector, d is an M-dimensional
! data vector and A is an M x N dimensional real matrix
! of maximum rank.
!
! ls_gmres implements non-symmetric GMRES method.
!
! 2012 Stanford Exploration Project
! Musa Maharramov, musa@sep.stanford.edu
!
module gmres
use error
use base_vector
use base_solver
implicit none
!
! IMPLEMENTATION
!
! tolerance/round-up
real, public, parameter ::  DEFAULT_LS_TOL = 1.e-9
!
!
! a sample  solver implementation of the
! Gemeralised Residuals Method (GMRES)
!
type, public, extends(ls_solver)  :: ls_gmres
! auxiliary CR structures:
  class(ls_vector),private,pointer :: p,z,w
!
! flag indicating a breakdown of Krylov iterations
  logical :: breakdown = .false.
! tolerance
  real :: ls_tol
contains
        procedure, public :: init => gmres_init
        procedure, public :: step => gmres_step
        procedure, public :: get_iter => gmres_getiter
        procedure, public :: stalled => gmres_stalled
end type ls_gmres
!
contains
!
!
! check if the iterations have stalled
  function gmres_stalled(this)
  logical :: gmres_stalled
  class(ls_gmres), intent(in) :: this
    gmres_stalled=this%breakdown
    return
  end function gmres_stalled
!
! return iteration count
  function gmres_getiter(this)
  integer :: gmres_getiter
  class(ls_gmres), intent(in) :: this
    gmres_getiter=this%iter
    return
  end function gmres_getiter

! ls_gmres solver initialisation
  subroutine gmres_init(this, op, m, d, resume, w, tol)
!
!   containing object
    class(ls_gmres), intent(inout) :: this

!   pointer to a procedure implementing A and A*
    class(ls_oper), intent(in), target :: op

!   initial model (m) and data d (right-hand side of (*))
!   have to be inout for error logging
    class(ls_vector), intent(inout) :: m, d
!
!   optional LS weights
    class(ls_vector), optional, intent(inout) :: w
!
!   resume flag - don't clear the state if .true.
    logical, optional :: resume
!
!   tolerance param
    real, optional, intent(in) :: tol
!
    if (.not.present(resume) .or. .not.resume) then
      call set_error(LS_OK); this%iter=0
    end if

    if (present(tol)) then
      this%ls_tol=tol
    else
      this%ls_tol=DEFAULT_LS_TOL
    end if

!   pointer to operator
    this%opr00=>op
!
!   init approx
    call m%spawn(this%m)
!   RHS
    call d%spawn(this%d)
!
!   create the residual vector - will form new (r,m)=x
    call d%spawn(this%r);
!   auxiliary vectors p, z, w
    call m%spawn(this%p); call m%spawn(this%z) 
    call d%spawn(this%w)
!   r=d-Fm
    call this%opr00%apply(this%m,this%r,.false.)
    call this%r%lc(this%d,this%r,1.,-1.)
!   z=FT r
    call this%opr00%apply(this%z,this%r,.true.)
    this%p=this%z
!
    this%breakdown = .false.
!    
    return
!
  end subroutine gmres_init

! ls_gmres solver step 
  subroutine gmres_step(this)
!   containing object
    class(ls_gmres), intent(inout) :: this
!   aux vars
    real :: alpha, beta, temp

    this%iter=this%iter+1

!   w=Ap
    call this%opr00%apply(this%p,this%w,.false.)
!   |w|^2
    temp=this%w%dot(this%w)
    if (abs(temp)<this%ls_tol) then
      this%breakdown=.true.
      return
    end if
!
    alpha=this%z%dot(this%z)/temp
!   r=r-alpha * w
    call this%r%lc(this%r,this%w,1.,-alpha)
!   m=m+alpha * p
    call this%m%lc(this%m,this%p,1.,alpha)
!   |z_i|^2
    temp=this%z%dot(this%z)
    if (abs(temp)<this%ls_tol) then
      this%breakdown=.true.
      return
    end if
!
!   z=FT r
    call this%opr00%apply(this%z,this%r,.true.)
!
!   beta=|z_{i+1}|^2/|z_i|^2
    beta=this%z%dot(this%z)/temp
!   p=z+beta * p
    call this%p%lc(this%z,this%p,1.,beta)
!
    return
  end subroutine gmres_step
!
end module gmres
