!
!  Simple vector implementation for the Generic Least 
!  Squares Solver abstract interface library.
!
! 2012 Stanford Exploration Project
! Musa Maharramov, musa@sep.stanford.edu
!
module simple_vector 
use error
use base_vector
implicit none
!
! IMPLEMENTATION
!
! An implementation of a vector - extends the abstract 'ls_vector' type
type, public, extends(ls_vector) :: vector
 real, allocatable, dimension (:), public :: val
 integer, public :: stat 
 integer, public :: n = 0
!
contains        
        procedure, public :: inobject => vec_inobject
        procedure, public :: outobject => vec_outobject
        procedure, public :: inarray => vec_inarray
        procedure, public :: outarray => vec_outarray
        procedure, public :: dot => vec_dot
        procedure, public :: lc => vec_lc
        procedure, public :: cwmult => vec_cwmult
        procedure, public :: spawn => vec_spawn
        procedure, public :: zero => vec_zero
!       automatic destructor        
        final :: free
end type vector
!
!
contains
!
! copy array vec to the vector
subroutine vec_inarray(this, vec, cnt)
class(vector), intent(inout) :: this
real, dimension(:), intent(in) :: vec
integer, intent(in) :: cnt
integer :: i
logical :: aflag
  aflag=allocated(this%val);
  if (.not.aflag .or. this%n/=cnt) then ! reallocate memory 
    this%n=0; 
    if (aflag) then
      deallocate(this%val, stat=i);
      if (i>0) then 
        call set_error(LS_MEM_FREE_ER)
        return
      end if
    end if
    allocate(this%val(1:cnt),stat=i)
    if (i>0) then 
      call set_error(LS_MEM_ALLOC_ER)
      return
    end if
    this%n=cnt
  end if
  this%val(1:this%n)=vec(1:this%n)
  call set_error(LS_OK)
  return
end subroutine vec_inarray

! copy one vector into another vec to the vector
subroutine vec_inobject(this, src)
class(vector), intent(inout) :: this
class(ls_vector), intent(in) ::src 
integer :: i,cnt
logical :: aflag

  select type(src)
  class is (vector) ! copy only if the same or derived type
    aflag=allocated(this%val);cnt=src%n
    if (cnt == 0) then
      call set_error(LS_UNINIT_ER)
      return
    end if
    if (.not.aflag .or. this%n/=cnt) then ! reallocate memory 
      this%n=0; 
      if (aflag) then
        deallocate(this%val, stat=i);
        if (i>0) then 
          call set_error(LS_MEM_FREE_ER)
          return
        end if
      end if
      allocate(this%val(1:cnt),stat=i)
      if (i>0) then 
        call set_error(LS_MEM_ALLOC_ER)
        return
      end if
      this%n=cnt
    end if
    this%val(1:cnt)=src%val(1:cnt)
    call set_error(LS_OK)
  class default 
    call set_error(LS_TYPE_MISMATCH_ER)
  end select
  return
end subroutine vec_inobject

subroutine vec_outarray(this, vec, cnt)
! no internal error logging for outarray
class(vector), intent(in) :: this
real, dimension(:), intent(inout) :: vec
integer, intent(in) :: cnt
  if (.not.allocated(this%val)) then
!     call set_error(LS_UNINIT_ER)
    return
  end if
  if (this%n/=cnt) then
!    call set_error(LS_SIZE_MISMATCH_ER)
    return
  end if
  vec(1:cnt)=this%val(1:cnt)
!  call set_error(LS_OK)
  return
end subroutine vec_outarray

subroutine vec_outobject(this, dest)
! no internal error logging for outobject
class(vector), intent(in) :: this
class(ls_vector), intent(inout) :: dest
  select type (dest)
  class is (vector)
    call dest%in(this)
  class default
!    can't log error, just return
!    call set_error(LS_TYPE_MISMATCH_ER)
    return
  end select
  return
end subroutine vec_outobject


subroutine vec_cwmult(this, u, v)
! has to be inout for error logging
class(vector), intent(inout) :: this
class(ls_vector), intent(in) :: u, v
integer :: i

  select type (u)
  class is (vector)
  select type(v)
  class is (vector)
    if (.not.allocated(this%val) .or. .not.allocated(u%val) .or. .not. allocated(v%val)) then
      call set_error(LS_UNINIT_ER)
      return
    end if
    do i=1,this%n
    end do
  class default
    call set_error(LS_TYPE_MISMATCH_ER)
    return
  end select
  class default
    call set_error(LS_TYPE_MISMATCH_ER)
    return
  end select
  call set_error(LS_OK)
  return
end subroutine vec_cwmult 


function vec_dot(this, vec)
! has to be inout for error logging
class(vector), intent(inout) :: this
class(ls_vector), intent(in) :: vec
real :: vec_dot
  vec_dot=0.
  select type (vec)
  class is (vector)
    if (.not.allocated(this%val) .or. .not.allocated(vec%val)) then
      call set_error(LS_UNINIT_ER)
      return
    end if
    vec_dot=dot_product(this%val, vec%val)
  class default
    call set_error(LS_TYPE_MISMATCH_ER)
    return
  end select
  call set_error(LS_OK)
  return
end function vec_dot

subroutine vec_spawn(this, res)
! has to be inout for error logging
class(vector), intent(inout) :: this
class(ls_vector), intent(out), pointer :: res
type(vector), pointer :: r
integer :: i
  if (.not.allocated(this%val)) then
    call set_error(LS_UNINIT_ER)
    return
  end if
  allocate(r, stat=i)
  if (i>0) then 
    call set_error(LS_MEM_ALLOC_ER)
    return
  end if
  r=this; res => r
  select type(res)
  class is (vector)
  if (allocated(res%val)) then
    res%stat=LS_OK           
  else
    res%stat=LS_UNINIT_ER  
  end if
  end select
  return
end subroutine vec_spawn


subroutine vec_lc(this, u, v, alpha, beta)
class(vector), intent(inout) :: this
class(ls_vector), intent(in) :: u
class(ls_vector), intent(in), optional :: v
real, intent(in), optional :: alpha, beta
logical :: aflag
integer :: i,cnt, cntv
  aflag=allocated(this%val)
! =alpha*u
  select type (u)
    class is (vector)
      cnt=u%n
      if (.not.allocated(u%val)) then
        call set_error(LS_UNINIT_ER)
        return
      end if
      if (.not.aflag .or. this%n/=cnt) then ! reallocate memory 
        this%n=0; 
        if (aflag) then
          deallocate(this%val, stat=i);
          if (i>0) then 
            call set_error(LS_MEM_FREE_ER)
            return
          end if
        end if
        allocate(this%val(1:cnt),stat=i)
        if (i>0) then 
          call set_error(LS_MEM_ALLOC_ER)
          return
        end if
        this%n=cnt
      end if
!     + beta*v
      if (present(v)) then
        select type (v)
        class is (vector)
          if (.not.allocated(v%val)) then
            call set_error(LS_UNINIT_ER)
            return
          end if
          cntv=v%n
          if (cnt /= cntv) then
            call set_error(LS_SIZE_MISMATCH_ER)
            return
          end if
          if (present(alpha) .and. present(beta)) then
            this%val=alpha*u%val+beta*v%val
          else if (present(alpha) .and. .not.present(beta)) then
            this%val=alpha*u%val+v%val
          else  if (.not.present(alpha) .and. present(beta)) then
            this%val=u%val+beta*v%val
          else
            this%val=u%val+v%val
          end if
        class default
          call set_error(LS_TYPE_MISMATCH_ER)
          return
        end select
      else
        if (present(alpha)) then
          this%val=alpha*u%val
        else 
          this%val=u%val
        end if
      end if
    class default
      call set_error(LS_TYPE_MISMATCH_ER)
      return
  end select
  call set_error(LS_OK)
end subroutine vec_lc

subroutine vec_zero(this)
class(vector), intent(inout) :: this
  if (allocated(this%val)) then
    this%val=0
  end if
  return
end subroutine vec_zero

! finalisation (clean-up) routine for the `vector` type
subroutine free(this)
type(vector), intent(inout) :: this
integer :: i
  if (allocated(this%val)) then
    deallocate(this%val, stat=i)
    if (i>0) then 
      call set_error(LS_MEM_FREE_ER)
    else
      call set_error(LS_OK)
    end if
  end if
  this%n=0
  write  (*,*) 'DBG: VECTOR DESTRUCTOR INVOKED'
  return
end subroutine free
!

end module simple_vector
