%
% Synthetic Reservoir model and pore pressure
% synthetic input generator for 'zpemodel.x' 0.5
%
% 2012 Musa Maharramov
% Stanford Exploration Project
% Stanford University
%
clc
%
% output file names
lacqrealfile='lacq_subsidence_real.bin';
ksatfile='ksat.bin';
kminfile='kmin.bin';
shearfile='shear.bin';
depthshearfile='depthshear.bin';
poissonfile='poisson.bin';
bottomsurffile='rzhi.bin';
bottomsurffileh='rzhi_hires.bin';
topsurffile='rzlo.bin';
topsurffileh='rzlo_hires.bin';
axisymppfile='symmetricpp.bin';
axisymppfileh='symmetricpp_hires.bin';
asymppfile='asymmetricpp.bin';
asymppfileh='asymmetricpp_hires.bin';
lacqheaderfile='lacq.hdr';
lacqheaderfileh='lacq_hires.hdr';
lacqheaderfiled='lacq_depth.hdr';
lacqheaderfilehd='lacq_hires_depth.hdr';
%
% all diatances in km
% moduli in GPa
% pressures in MPa
%
% model parameters
nx=101;ny=101;nz=20;
x0=-10.;y0=-10.;z0=0;
dx=-2.*x0/(nx-1);dy=-2.*y0/(ny-1);dz=.1;
% low-res reservoir grid
rnx=21;rny=21;
% resrevoir radius
radius=8.;
rx0=-radius;ry0=-radius;
rdx=2.*radius/(rnx-1);rdy=2.*radius/(rny-1);
%
% hi-res reservoir grid
rnxh=81;rnyh=81;
% resrevoir radius
radius=8.;
rx0=-radius;ry0=-radius;
rdxh=2.*radius/(rnxh-1);rdyh=2.*radius/(rnyh-1);
%
% First, extrapolate 1D data from Segall et.al
% to a full 3D subsidence profile.
%
% Paul Segall, Jean-Robert Grasso, and Antony
% Mossop. Poroelastic stressing and induced
% seismicity near the Lacq gas field, 
% southwestern France. Journal of Geophysics
% Research, 99:15423–15438, 1994.
% http://dx.doi.org/10.1029/94JB00989.
%
% south-east subsidence profile (m)
se_subs = [58 53 30 0 0]/1000;
% se dist (km)
se_dist = [0 3 4 9 13];
% north-west subsidence profile (m)
nw_subs = [58 23 20 19 0]/1000;
% noerg-est dist (km)
nw_dist = [0 4  6 10 13];
%
dense_radii=linspace(0.,13.,100);
%
se_profile=interp1(se_dist,se_subs,dense_radii,'linear');
nw_profile=interp1(se_dist,nw_subs,dense_radii,'linear');
%
% presumed lacq subsidence in polar coordinates  
dense_phi=linspace(0.,2*pi,100);
lacq_subs_polar=zeros(100,100);
for iphi=1:100
  phi=(iphi-1)*2*pi/100.;
  for irho=1:100    
    lacq_subs_polar(irho,iphi)=nw_profile(irho)+...
         (se_profile(irho)-nw_profile(irho))*sin((phi+pi/4)/2.)^2;
  end
end
% convert to cartesian
lacq_real=zeros(3*nx,ny);
zlac_real=zeros(nx,ny)
for iy=1:ny
  y=dy*(iy-1)+y0;
  for ix=1:nx
    x=dx*(ix-1)+x0;
    iphi=min(floor((atan2(y,x)+pi)/(2.*pi/100.))+1,100);
    irho=min(floor(((x^2+y^2)^.5)/(13./100.))+1,100);
    lacq_real(3*ix-2:3*ix-1,iy)=0;
    lacq_real(3*ix,iy)=lacq_subs_polar(irho,iphi);
    zlacq_real(ix,iy)=lacq_subs_polar(irho,iphi);
  end
end
%
% write simulated 'real' subsidence daya to binary file
%
hfile=fopen(lacqrealfile,'w','ieee-le');
fwrite(hfile,lacq_real,'float');
fclose(hfile);
    
%
% reservoir bounding surfaces
rzlo=zeros(rnx,rny);
rzhi=zeros(rnx,rny);
% specified pore pressure drop
pp=zeros(rnx,rny);
pph=zeros(rnxh,rnyh);
ppasym=pp;
ppasymhh=pph;
% amplitude of the reservoir dome
A=1.4;
% characteristic length of structure
lc=5.;
% depth of the reservoir
d0=3.5;
% max pressure drop (MPa)
pmax=55;
% reservoir width
rwidth=.35;
% peak shifted
sx0=3.;sy0=3.;
for iy=1:rny
  for ix=1:rnx    
    rho=((rx0+(ix-1)*rdx)^2+...
        (ry0+(iy-1)*rdy)^2)^.5;
    rho2=((rx0+(ix-1)*rdx-sx0)^2+...
        (ry0+(iy-1)*rdy-sy0)^2)^.5;
    rzlo(ix,iy)=d0-A*(exp(-(rho/lc)^2)-.5);
    rzhi(ix,iy)=rzlo(ix,iy)+rwidth;
    pp(ix,iy)=pmax*exp(-(rho/...
        radius)^4);
    ppasym(ix,iy)=pmax*exp(-(rho2/...
        (radius+(sin(atan2(ry0+(iy-1)*rdy-sy0,rx0+(ix-1)*rdx-sx0)-pi/4)^2)*radius/3))^4);
% make the reservois circular        
    if (rho > radius)
      rzhi(ix,iy)=rzlo(ix,iy);
    end    
  end
end
%
for iy=1:rnyh
  for ix=1:rnxh
    rho=((rx0+(ix-1)*rdxh)^2+...
        (ry0+(iy-1)*rdyh)^2)^.5;
    rho2=((rx0+(ix-1)*rdxh-sx0)^2+...
        (ry0+(iy-1)*rdyh-sy0)^2)^.5;
    rzloh(ix,iy)=d0-A*(exp(-(rho/lc)^2)-.5);
    rzhih(ix,iy)=rzloh(ix,iy)+rwidth;
    pph(ix,iy)=pmax*exp(-(rho/...
        radius)^4);
    ppasymh(ix,iy)=pmax*exp(-(rho2/...
        (radius+(sin(atan2(ry0+(iy-1)*rdyh-sy0,rx0+(ix-1)*rdxh-sx0)-pi/4)^2)*radius/3))^4);
% make the reservois circular        
    if (rho > radius)
      rzhih(ix,iy)=rzloh(ix,iy)-0.1;
    end    
  end
end

% reservoir bounding surfaces 
% to binary files
% write
hfile=fopen(topsurffile,'w','ieee-le');
fwrite(hfile,rzlo,'float');
fclose(hfile);
hfile=fopen(bottomsurffile,'w','ieee-le');
fwrite(hfile,rzhi,'float');
fclose(hfile);
% ditto high-res
hfile=fopen(topsurffileh,'w','ieee-le');
fwrite(hfile,rzloh,'float');
fclose(hfile);
hfile=fopen(bottomsurffileh,'w','ieee-le');
fwrite(hfile,rzhih,'float');
fclose(hfile);
%
% write synthetic pore pressures
% to binary files
%
hfile=fopen(axisymppfile,'w','ieee-le');
fwrite(hfile,pp,'float');
fclose(hfile);
hfile=fopen(asymppfile,'w','ieee-le');
fwrite(hfile,ppasym,'float');
fclose(hfile);
% ditto high-res
hfile=fopen(axisymppfileh,'w','ieee-le');
fwrite(hfile,pph,'float');
fclose(hfile);
hfile=fopen(asymppfileh,'w','ieee-le');
fwrite(hfile,ppasymh,'float');
fclose(hfile);
%
% init poroelastic medium parameters
% and write to binary files
%
% we only care about ksat/kmin
% so the value of the undrained
% and mineral moduli are relative;
% undrained bulk modulus
model=.75*ones(nz,nx,ny); 
% write
hfile=fopen(ksatfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);
% mineral grain bulk modulus
model=ones(nz,nx,ny); ;
% write
hfile=fopen(kminfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);
% shear modulus (in GPa)
model=23.*ones(nz,nx,ny);
% write
hfile=fopen(shearfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);

% asymmetric shear modulus - scale with depth
for iz=1:nz
  model(iz,:,:)=model(iz,:,:)*1./(1.+(nz-iz)/nz);
end
% write
hfile=fopen(depthshearfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);
% Poisson's ratio
model=.25*ones(nz,nx,ny); ;
% write
hfile=fopen(poissonfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);
%
% create reservoir header 
% readable by expt_tk_0.2
% Fortran 2003 Object-oriented
% Framework
%
hfile=fopen(lacqheaderfile,'w');
fprintf(hfile,'nx=%d\n', nx);
fprintf(hfile,'ny=%d\n', ny);
fprintf(hfile,'nz=%d\n', nz);
fprintf(hfile,'x0=%f\n', x0);
fprintf(hfile,'y0=%f\n', y0);
fprintf(hfile,'z0=%f\n', z0);
fprintf(hfile,'dx=%f\n', dx);
fprintf(hfile,'dy=%f\n', dy);
fprintf(hfile,'dz=%f\n', dz);
%
fprintf(hfile,'rnx=%d\n', rnx);
fprintf(hfile,'rny=%d\n', rny);
fprintf(hfile,'rx0=%f\n', rx0);
fprintf(hfile,'ry0=%f\n', ry0);
fprintf(hfile,'rdx=%f\n', rdx);
fprintf(hfile,'rdy=%f\n', rdy);
%
fprintf(hfile,'real rzlo(rnx,rny)=%s\n', topsurffile);
fprintf(hfile,'real rzhi(rnx,rny)=%s\n', bottomsurffile);
fprintf(hfile,'real ksat(nz,nx,ny)=%s\n', ksatfile);
fprintf(hfile,'real kmin(nz,nx,ny)=%s\n', kminfile);
fprintf(hfile,'real mu(nz,nx,ny)=%s\n', shearfile);
fprintf(hfile,'real nu(nz,nx,ny)=%s\n', poissonfile);
%
fclose(hfile);
%
% ditto high-res
hfile=fopen(lacqheaderfileh,'w');
fprintf(hfile,'nx=%d\n', nx);
fprintf(hfile,'ny=%d\n', ny);
fprintf(hfile,'nz=%d\n', nz);
fprintf(hfile,'x0=%f\n', x0);
fprintf(hfile,'y0=%f\n', y0);
fprintf(hfile,'z0=%f\n', z0);
fprintf(hfile,'dx=%f\n', dx);
fprintf(hfile,'dy=%f\n', dy);
fprintf(hfile,'dz=%f\n', dz);
%
fprintf(hfile,'rnx=%d\n', rnxh);
fprintf(hfile,'rny=%d\n', rnyh);
fprintf(hfile,'rx0=%f\n', rx0);
fprintf(hfile,'ry0=%f\n', ry0);
fprintf(hfile,'rdx=%f\n', rdxh);
fprintf(hfile,'rdy=%f\n', rdyh);
%
fprintf(hfile,'real rzlo(rnx,rny)=%s\n', topsurffileh);
fprintf(hfile,'real rzhi(rnx,rny)=%s\n', bottomsurffileh);
fprintf(hfile,'real ksat(nz,nx,ny)=%s\n', ksatfile);
fprintf(hfile,'real kmin(nz,nx,ny)=%s\n', kminfile);
fprintf(hfile,'real mu(nz,nx,ny)=%s\n', shearfile);
fprintf(hfile,'real nu(nz,nx,ny)=%s\n', poissonfile);
%
fclose(hfile);


% ditto for depth-dependentparameters
hfile=fopen(lacqheaderfiled,'w');
fprintf(hfile,'nx=%d\n', nx);
fprintf(hfile,'ny=%d\n', ny);
fprintf(hfile,'nz=%d\n', nz);
fprintf(hfile,'x0=%f\n', x0);
fprintf(hfile,'y0=%f\n', y0);
fprintf(hfile,'z0=%f\n', z0);
fprintf(hfile,'dx=%f\n', dx);
fprintf(hfile,'dy=%f\n', dy);
fprintf(hfile,'dz=%f\n', dz);
%
fprintf(hfile,'rnx=%d\n', rnx);
fprintf(hfile,'rny=%d\n', rny);
fprintf(hfile,'rx0=%f\n', rx0);
fprintf(hfile,'ry0=%f\n', ry0);
fprintf(hfile,'rdx=%f\n', rdx);
fprintf(hfile,'rdy=%f\n', rdy);
%
fprintf(hfile,'real rzlo(rnx,rny)=%s\n', topsurffile);
fprintf(hfile,'real rzhi(rnx,rny)=%s\n', bottomsurffile);
fprintf(hfile,'real ksat(nz,nx,ny)=%s\n', ksatfile);
fprintf(hfile,'real kmin(nz,nx,ny)=%s\n', kminfile);
fprintf(hfile,'real mu(nz,nx,ny)=%s\n', depthshearfile);
fprintf(hfile,'real nu(nz,nx,ny)=%s\n', poissonfile);
%
fclose(hfile);
%
% ditto high-res
hfile=fopen(lacqheaderfilehd,'w');
fprintf(hfile,'nx=%d\n', nx);
fprintf(hfile,'ny=%d\n', ny);
fprintf(hfile,'nz=%d\n', nz);
fprintf(hfile,'x0=%f\n', x0);
fprintf(hfile,'y0=%f\n', y0);
fprintf(hfile,'z0=%f\n', z0);
fprintf(hfile,'dx=%f\n', dx);
fprintf(hfile,'dy=%f\n', dy);
fprintf(hfile,'dz=%f\n', dz);
%
fprintf(hfile,'rnx=%d\n', rnxh);
fprintf(hfile,'rny=%d\n', rnyh);
fprintf(hfile,'rx0=%f\n', rx0);
fprintf(hfile,'ry0=%f\n', ry0);
fprintf(hfile,'rdx=%f\n', rdxh);
fprintf(hfile,'rdy=%f\n', rdyh);
%
fprintf(hfile,'real rzlo(rnx,rny)=%s\n', topsurffileh);
fprintf(hfile,'real rzhi(rnx,rny)=%s\n', bottomsurffileh);
fprintf(hfile,'real ksat(nz,nx,ny)=%s\n', ksatfile);
fprintf(hfile,'real kmin(nz,nx,ny)=%s\n', kminfile);
fprintf(hfile,'real mu(nz,nx,ny)=%s\n', depthshearfile);
fprintf(hfile,'real nu(nz,nx,ny)=%s\n', poissonfile);
%
fclose(hfile);


% model plotting grid
[ccx, ccy]=meshgrid(linspace(x0,x0+(nx-1)*dx,nx),...
    linspace(y0,y0+(ny-1)*dy,ny));
%
figure
    surf(ccx, ccy, zlacq_real);
    xlabel('x - km');
    ylabel('y - km');
    zlabel('m');
    set(gca,'ZDir','reverse');
    title('Presumed Vert Displacement at Lacq');
    shading interp;
    colorbar
figure
    [C,h]=contour(ccx, ccy, zlacq_real);
    clabel(C,h);
    xlabel('x - km');
    ylabel('y - km');
    zlabel('m');
    set(gca,'ZDir','reverse');
    title('Presumed Vert Displacement at Lacq in metres');
    shading interp;
    colorbar
%
% reservoir plotting grid
[cx, cy]=meshgrid(linspace(rx0,rx0+rdx*(rnx-1),rnx),...
    linspace(ry0,ry0+rdy*(rnx-1),rny));
figure
    surf(cx,cy,pp);
    title('Synthetic Axisymmetric Pressure Drop MPa');
    xlabel('x - km');
    ylabel('y - km');
    zlabel('MPa')
    shading interp;
    colorbar
figure
    [C,h]=contour(cx,cy,pp);
    clabel(C,h);
    title('Synthetic Axisymmetric Pressure Drop MPa');
    xlabel('x - km');
    ylabel('y - km');
    zlabel('MPa')
    shading interp;
    colorbar

figure
    surf(cx,cy,ppasym);
    title('Synthetic Asymmetric Pressure Drop MPa');
    xlabel('x - km');
    ylabel('y - km');
    shading interp;
    colorbar
figure
    [C,h]=contour(cx,cy,ppasym);
    clabel(C,h);
    title('Synthetic Asymmetric Pressure Drop MPa');
    xlabel('x - km');
    ylabel('y - km');
    zlabel('MPa')
    shading interp;
    colorbar
  
figure
    surf(cx,cy,rzlo);
    set(gca,'ZDir','reverse');
    xlabel('x - km');
    ylabel('y - km');
    title('Reservoir Dome');
    shading interp;
    colorbar
