! Modelling deformation due to 
! pore pressure change
! in a poroelastostatic medium
!
! 2012 Musa Maharramov, Stanford University
!
! Version 0.5
! layered elastic medium -- displacemnent modelling only
!
module zpemodel_defs
use poroelastic_depth
use simple_vector
use io
use omp_lib

implicit none

integer, parameter, public :: MAXLEN = 1024
integer, parameter, public :: MAXARG = 10

! info and error strings
character(len=*),parameter,public :: MSG_GREET= &
  'Z-PEMODEL - MODELLING LAYERED EARTH DEFORMATION DUE TO PORE PRESSURE CHANGE' 
character(len=*),parameter,public :: MSG_USAGE= &
'zpemodel.x par=<Parameter header>   &
 Parameter header: &
   model={0|1}[1] &
   pore=<binary file> &
   displacement=<binary file> &
   reservoir=<Reservoir header> &
   dispout=<file name> &  
   poreout=<file name> &
   maxiter=<num>[100] &
   mx={0|1}[1] &
   my={0|1}[1] &
   mz={0|1}[1] &
   nxinit=<int>[res%rnx] &
   nyinit=<int>[res%rny] &
   dxinit=<real>[res%rdx] &
   dyinit=<real>[res%rdy] &
 Reservoir header: &
   nx=<int> &
   ny=<int> &
   nz=<int> &
   dx=<real> &
   dy=<real> &
   dz=<real> &
   x0=<real> &
   y0=<real> &
   z0=<real> &
   rnx=<int> &
   rny=<int> &
   rdx=<real> &
   rdy=<real> &
   rx0=<real> &
   ry0=<real> &
   real rzlo(rnx,rny)=<binary file> &
   real rzhi(rnx,rny)=<binary file>&
   real ksat/kmin/shear/poisson(nz,nx,ny)=<binary files>'

character(len=*),parameter,public :: MSG_INIT='INITING THE OPERATOR'
character(len=*),parameter,public :: MSG_RUNINV='INVERSION RUN'
character(len=*),parameter,public :: MSG_RUNMOD='MODELLING RUN'
character(len=*),parameter,public :: MSG_WRT='WRITING OUTPUT'
character(len=*),parameter,public :: MSG_FIN='FINISHED'

! application parameters
type, extends(persistent) :: param_type
! model (1) or invert flag
integer :: model
! is initial approx specified (1) or not
integer :: init
! max no of solver iterations
! in the inversion mode
integer :: maxiter
!
! no of OMP threads
integer, public :: np

! component flags
integer :: mx, my, mz
!
! iteration count
integer :: cnt
!
! pore pressure and displacement
class(vector), pointer :: pp, disp
! optional initial approximation 
! to the pore-pressure drop
! in the inversion mode
class(vector), pointer :: ppinit
! grid parameters for the initial
! approximation (note x0init and y0init
! are assumed equal to rx0 and ry0
! reservoir model
integer :: nxinit, nyinit
real :: dxinit, dyinit
class(base_reservoir), pointer :: res
!
! reservoir header name
character(len=IO_STRLEN) :: resheader
! displacement and pore-pressure drop output 
! binary file names
character(len=IO_STRLEN) :: dispout, poreout
contains
      procedure :: load => lpar
      procedure :: store => spar
end type param_type

! default max iteration count 
! for the inverse solver
integer, parameter :: DEFMAXITER = 100
contains

! load params from header
subroutine lpar(this,header)
class(param_type), intent(inout) :: this
class(header_type), intent(inout) :: header
type(header_type) :: resh
! aux
integer :: m
  call header%load(this%model,'model',1)
  call header%load(this%np,'np',1)
  call header%load(this%init,'init',0)
  call header%load(this%maxiter,'maxiter',DEFMAXITER)
  call header%load(this%mx,'mx',1)
  call header%load(this%my,'my',1)
  call header%load(this%mz,'mz',1)
  call header%load(this%dispout,'dispout','dispout')
  call header%load(this%poreout,'poreout','poreout')
  if (associated(this%res)) nullify(this%res)

! allocate reservoir model
  allocate(pe_reservoir::this%res)
  call header%load(this%resheader,'reservoir','reservoir')
  call resh%in(this%resheader)
  call this%res%load(resh)
  call resh%close()
! allocate pore-pressure drop
! and displacement fields
  m=this%res%rnx*this%res%rny;allocate(this%pp)
  this%pp%n=m;allocate(this%pp%val(m))
!
  m=3*this%res%nx*this%res%ny*this%res%nz
  allocate(this%disp)
  this%disp%n=m;allocate(this%disp%val(m))
!
  if (this%model==1) then
    call header%load(this%pp%val,'pore')
  else
    call header%load(this%disp%val,'displacement')
  end if
!
! load the initial approximation to the pore-pressure
! drop (e.g., defined on a sparser grid)
  if (this%model==0 .and. header%defined('poreinit')) then
    call header%load(this%nxinit,'nxinit',this%res%rnx)
    call header%load(this%nyinit,'nyinit',this%res%rny)
    call header%load(this%dxinit,'dxinit',this%res%rdx)
    call header%load(this%dyinit,'dyinit',this%res%rdy)
    m=this%nxinit*this%nyinit;allocate(this%ppinit)
    this%ppinit%n=m;allocate(this%ppinit%val(m))
!
    call header%load(this%ppinit%val,'poreinit')
  end if
end subroutine lpar

! save params to header
subroutine spar(this,header)
class(param_type), intent(inout) :: this
class(header_type), intent(inout) :: header
type(header_type) :: resh
  call header%store(this%model,'model')
  call header%store(this%np,'np')
  call header%store(this%init,'init')
  call header%store(this%maxiter,'maxiter')
  call header%store(this%mx,'mx')
  call header%store(this%my,'my')
  call header%store(this%mz,'mz')
  call header%store(this%cnt,'iters')
  call header%store(this%resheader,'reservoir')
  call header%store(this%dispout,'dispout')
  call header%store(this%poreout,'poreout')

! no need to write reservoir model as we didn't change it
  if (this%model==1) then
    call header%store(this%res%nz, 'nz')
    call header%store(this%res%nx, 'nx')
    call header%store(this%res%ny, 'ny')
    call header%store(this%res%dx, 'dx')
    call header%store(this%res%dy, 'dy')
    call header%store(this%res%dz, 'dz')
    call header%store(this%res%x0, 'x0')
    call header%store(this%res%y0, 'y0')
    call header%store(this%res%z0, 'z0')
    call header%store(this%disp%val, &
    'real displacement(3,nz,nx,ny)='//& 
      trim(this%dispout))
  else
    call header%store(this%res%rnx, 'rnx')
    call header%store(this%res%rny, 'rny')
    call header%store(this%res%rdx, 'rdx')
    call header%store(this%res%rdy, 'rdy')
    call header%store(this%res%rx0, 'rx0')
    call header%store(this%res%ry0, 'ry0')
    call header%store(this%pp%val,'real pore(rnx, rny)='// &
      trim(this%poreout)) 
  end if
!
end subroutine spar

! linearly interpolate (initial approximation to)
! pore-pressure drop 'pcoarse' from grid 
! (nx1,ny1,dx1,dy1) onto grid (nx,ny,dx,dy)
subroutine interp_pp(pcoarse, nx1, ny1, dx1, dy1, &
  pdense, nx, ny, dx, dy)
real, dimension (:), intent(inout) :: pcoarse
real, dimension (:), intent(inout) :: pdense
integer, intent(in) :: nx, ny, nx1, ny1
real, intent(in) :: dx, dy, dx1, dy1
! aux
integer :: i, j, i1, i2, j1, j2
real :: wy, wx
  do j=1,ny
    j1=dy*real(j-1)/dy1+1
    wy=1.-(dy*real(j-1)-real(j1-1)*dy1)/dy1
    if (j1<ny1) then
      j2=j1+1
    else
      j2=j1
    end if
    do i=1,nx
      i1=dx*real(i-1)/dx1+1
      wx=1.-(dx*real(i-1)-real(i1-1)*dx1)/dx1
      if (i1<nx1) then
        i2=i1+1
      else
        i2=i1
      end if
      pdense(nx*(j-1)+i)= &
      pcoarse(nx1*(j1-1)+i1)*wy*wx+ &
      pcoarse(nx1*(j1-1)+i2)*wy*(1.-wx)+ &
      pcoarse(nx1*(j2-1)+i1)*(1.-wy)*wx+ &
      pcoarse(nx1*(j2-1)+i2)*(1.-wy)*(1.-wx)    
    end do
  end do
  return
end subroutine interp_pp

end module zpemodel_defs

