!
! Modelling deformation due to 
! pore pressure change
! in a poroelastostatic medium
!
! 2012 Musa Maharramov
! Stanford Exploration Project
!
! Version 0.4 using type(persistent)
! and optional initial approximation
! to inverted pore-pressure drop
! defined on a different (sparser)
! grid
!
! 1) In the modelling mode, the code
!    computes 3D displacement field 
!    for a specified poroelastic reservoir
!    model from a user-supplied
!    pore-pressure drop within the 
!    reservoir.
!
! 2) In the inversion mode, a 
!    user-specified 3D displacement field
!    (e.g. either subsidence from levelling
!    or subsurface displacements from
!    time-lapse seismic) is inverted to
!    produce estimated pore-pressure drop
!    within the reservoir.
!
! 3) This version make no use of SEP/SU/DDS
!    or any other data I/O libraries, 
!    just the standard F03 features; hence
!    this code is portable across
!    all compliant F03 platforms.
!
!
program pemodel
! F03 poroelastostaic modelling 
! object-oriented framework 
! is referenced from 
! within this auxiliary module
use pemodel_defs
! Krylov solver (can pick any)
! from F03 optimisation
! object-oriented framework
use cgnr
!
implicit none
!
! modelling operator
class(ls_oper), pointer :: op
!
! iterative least-squares solver
type(ls_cgnr) :: solver
!
! application parameters
type(param_type) :: par
!
! parameter header
type(header_type) :: header
!

write(0,*)  MSG_GREET
!
! allocate the operator object --
! poroelastostatic modelling
! operator
allocate(pe_green::op)
!
call set_error(IO_OK)
select type(op)
class is (pe_green)
!
  call header%in(getarg('parin'))
  if (get_error()/= IO_OK) then
    write (0,*) MSG_USAGE
    stop 9555
  end if
  call par%load(header)
  if (get_error()/= IO_OK) then
    write (0,*) MSG_USAGE
    stop 9555
  end if
  call header%close()
  !
!
! set up OpenMP
  call omp_set_num_threads(par%np)
!  
! initialiase the operator  
  write(0,*)  MSG_INIT
  call op%init(par%res,par%mx==1,par%my==1,par%mz==1)
  !
  if (par%model==1) then
!  modelling
    write(0,*)  MSG_RUNMOD
    call op%apply(m=par%pp,d=par%disp,adj=.false.)
  else
!   inversion          
    write(0,*)  MSG_RUNINV
!   if an initial approximation to the pore-pressure
!   drop is specified, copy (interpolate) it into 
!   par%pp
    if (associated(par%ppinit)) then
      call interp_pp(par%ppinit%val,par%nxinit, par%nyinit, &
                  par%dxinit, par%dyinit, &
                  par%pp%val, par%res%rnx, par%res%rny, &
                  par%res%rdx, par%res%rdy)
    else
!     otherwise use zero initial approximation
      call par%pp%zero()
    end if
!   initialise the solver    
    call solver%init(op,par%pp,par%disp,.true.,w=par%weights,tol=par%tol)
!   iterate
    do 
      par%cnt=solver%get_iter()
      if (par%cnt>par%maxiter .or. solver%stalled()) exit
      write(0,*) 'ITER ', par%cnt
      call solver%step()
    end do
    par%pp=solver%m
!   scale back
    par%pp%val=par%pp%val/1000.
  end if
  write (0,*) MSG_WRT
!
  call header%out(getarg('parout'))
  call par%store(header)
  call header%close()
class default
end select

write(0,*)  MSG_FIN
end program pemodel

