%
% Synthetic Reservoir model and pore pressure
% synthetic input generator for 'pemodel.x' 0.2
%
% 2012 Musa Maharramov
% Stanford Exploration Project
% Stanford University
%
clc
%
% output file names
lacqrealfile='lacq_subsidence_real.bin';
ksatfile='l_ksat.bin';
kminfile='l_kmin.bin';
shearfile='l_shear.bin';
poissonfile='l_poisson.bin';
bottomsurffile='l_rzhi.bin';
bottomsurffilem='l_rzhi_midres.bin';
bottomsurffileh='l_rzhi_hires.bin';
topsurffile='l_rzlo.bin';
topsurffilem='l_rzlo_midres.bin';
topsurffileh='l_rzlo_hires.bin';
lacqheaderfile='l_lacq.hdr';
lacqheaderfilem='l_lacq_midres.hdr';
lacqheaderfileh='l_lacq_hires.hdr';
%
% all diatances in km
% moduli in GPa
% pressures in MPa
%
% model parameters
nx=101;ny=101;
x0=-10.;y0=-10.;z0=0;
dx=-2.*x0/(nx-1);dy=-2.*y0/(ny-1);dz=.3;
% low-res reservoir grid
rnx=11;rny=11;
% resrevoir radius
radius=8.;
rx0=-radius;ry0=-radius;
rdx=2.*radius/(rnx-1);rdy=2.*radius/(rny-1);
%
% mid-res reservoir grid
rnxm=41;rnym=41;
rdxm=2.*radius/(rnxm-1);rdym=2.*radius/(rnym-1);
%
% hi-res reservoir grid
rnxh=81;rnyh=81;
rdxh=2.*radius/(rnxh-1);rdyh=2.*radius/(rnyh-1);
%
%
% First, extrapolate 1D data from Segall et.al
% to a full 3D subsidence profile.
%
% Paul Segall, Jean-Robert Grasso, and Antony
% Mossop. Poroelastic stressing and induced
% seismicity near the Lacq gas field, 
% southwestern France. Journal of Geophysics
% Research, 99:15423–15438, 1994.
% http://dx.doi.org/10.1029/94JB00989.
%
% south-east subsidence profile (m)
se_subs = [58 53 30 0 0]/1000;
% se dist (km)
se_dist = [0 3 4 9 13];
% north-west subsidence profile (m)
nw_subs = [58 23 20 19 0]/1000;
% noerg-est dist (km)
nw_dist = [0 4  6 10 13];
%
NRP=500;
dense_radii=linspace(0.,13.,NRP);
%
se_profile=interp1(se_dist,se_subs,dense_radii,'linear');
nw_profile=interp1(se_dist,nw_subs,dense_radii,'linear');
%
% presumed lacq subsidence in polar coordinates  
dense_phi=linspace(0.,2*pi,NRP);
lacq_subs_polar=zeros(NRP,NRP);
for iphi=1:NRP
  phi=(iphi-1)*2*pi/NRP;
  for irho=1:NRP    
    lacq_subs_polar(irho,iphi)=nw_profile(irho)+...
         (se_profile(irho)-nw_profile(irho))*sin((phi+pi/4)/2.)^2;
  end
end
% convert to cartesian
lacq_real=zeros(3*nx,ny);
zlac_real=zeros(nx,ny);
for iy=1:ny
  y=dy*(iy-1)+y0;
  for ix=1:nx
    x=dx*(ix-1)+x0;
    iphi=min(floor((atan2(y,x)+pi)/(2.*pi/NRP))+1,NRP);
    irho=min(floor(((x^2+y^2)^.5)/(13./NRP))+1,NRP);
    lacq_real(3*ix-2:3*ix-1,iy)=0;
    lacq_real(3*ix,iy)=lacq_subs_polar(irho,iphi);
    zlacq_real(ix,iy)=lacq_subs_polar(irho,iphi);
  end
end
%
% write simulated 'real' subsidence daya to binary file
%
hfile=fopen(lacqrealfile,'w','ieee-le');
fwrite(hfile,lacq_real,'float');
fclose(hfile);
    
%
% reservoir bounding surfaces
rzlo=zeros(rnx,rny);
rzhi=zeros(rnx,rny);
% amplitude of the reservoir dome
A=1.4;
% characteristic length of structure
lc=5.;
% depth of the reservoir
d0=3.5;
% max pressure drop (MPa)
pmax=55;
% reservoir width
rwidth=.3;
%
% low-res reservoir grid
%
for iy=1:rny
  for ix=1:rnx    
    rho=((rx0+(ix-1)*rdx)^2+...
        (ry0+(iy-1)*rdy)^2)^.5;    
    rzlo(ix,iy)=d0-A*(exp(-(rho/lc)^2)-.5);
    rzhi(ix,iy)=rzlo(ix,iy)+rwidth;    
% make the reservois circular        
    if (rho > radius)
      rzhi(ix,iy)=rzlo(ix,iy);
    end    
  end
end
%
% mid-resolution reservoir grid
%
rzlom=zeros(rnxm,rnym);
rzhim=zeros(rnxm,rnym);
for iy=1:rnym
  for ix=1:rnxm
    rho=((rx0+(ix-1)*rdxm)^2+...
        (ry0+(iy-1)*rdym)^2)^.5;
    
    rzlom(ix,iy)=d0-A*(exp(-(rho/lc)^2)-.5);
    rzhim(ix,iy)=rzlom(ix,iy)+rwidth;    
% make the reservois circular        
    if (rho > radius)
      rzhim(ix,iy)=rzlom(ix,iy)-0.1;
    end    
  end
end
%
% hi-res reservoir grid
%
rzloh=zeros(rnxh,rnyh);
rzhih=zeros(rnxh,rnyh);
for iy=1:rnyh
  for ix=1:rnxh
    rho=((rx0+(ix-1)*rdxh)^2+...
        (ry0+(iy-1)*rdyh)^2)^.5;    
    rzloh(ix,iy)=d0-A*(exp(-(rho/lc)^2)-.5);
    rzhih(ix,iy)=rzloh(ix,iy)+rwidth;    
% make the reservois circular        
    if (rho > radius)
      rzhih(ix,iy)=rzloh(ix,iy)-0.1;
    end    
  end
end
%
% reservoir bounding surfaces 
% to binary files
% write
hfile=fopen(topsurffile,'w','ieee-le');
fwrite(hfile,rzlo,'float');
fclose(hfile);
hfile=fopen(bottomsurffile,'w','ieee-le');
fwrite(hfile,rzhi,'float');
fclose(hfile);
% ditto mid-res
hfile=fopen(topsurffilem,'w','ieee-le');
fwrite(hfile,rzlom,'float');
fclose(hfile);
hfile=fopen(bottomsurffilem,'w','ieee-le');
fwrite(hfile,rzhim,'float');
fclose(hfile);
% ditto high-res
hfile=fopen(topsurffileh,'w','ieee-le');
fwrite(hfile,rzloh,'float');
fclose(hfile);
hfile=fopen(bottomsurffileh,'w','ieee-le');
fwrite(hfile,rzhih,'float');
fclose(hfile);
%
% init poroelastic medium parameters
% and write to binary files
%
% we only care about ksat/kmin
% so the value of the undrained
% and mineral moduli are relative;
% undrained bulk modulus
model=.75*ones(1,nx,ny); 
% write
hfile=fopen(ksatfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);
% mineral grain bulk modulus
model=ones(1,nx,ny); ;
% write
hfile=fopen(kminfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);
% shear modulus (in GPa)
model=23.*ones(1,nx,ny); ;
% write
hfile=fopen(shearfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);
% POisson's ratio
model=.25*ones(1,nx,ny); ;
% write
hfile=fopen(poissonfile,'w','ieee-le');
fwrite(hfile,model,'float');
fclose(hfile);
%
% create reservoir header 
% readable by expt_tk_0.2
% Fortran 2003 Object-oriented
% Framework
%
hfile=fopen(lacqheaderfile,'w');
fprintf(hfile,'nx=%d\n', nx);
fprintf(hfile,'ny=%d\n', ny);
fprintf(hfile,'nz=%d\n', 1);
fprintf(hfile,'x0=%f\n', x0);
fprintf(hfile,'y0=%f\n', y0);
fprintf(hfile,'z0=%f\n', z0);
fprintf(hfile,'dx=%f\n', dx);
fprintf(hfile,'dy=%f\n', dy);
fprintf(hfile,'dz=%f\n', dz);
%
fprintf(hfile,'rnx=%d\n', rnx);
fprintf(hfile,'rny=%d\n', rny);
fprintf(hfile,'rx0=%f\n', rx0);
fprintf(hfile,'ry0=%f\n', ry0);
fprintf(hfile,'rdx=%f\n', rdx);
fprintf(hfile,'rdy=%f\n', rdy);
%
fprintf(hfile,'real rzlo(rnx,rny)=%s\n', topsurffile);
fprintf(hfile,'real rzhi(rnx,rny)=%s\n', bottomsurffile);
fprintf(hfile,'real ksat(nz,nx,ny)=%s\n', ksatfile);
fprintf(hfile,'real kmin(nz,nx,ny)=%s\n', kminfile);
fprintf(hfile,'real mu(nz,nx,ny)=%s\n', shearfile);
fprintf(hfile,'real nu(nz,nx,ny)=%s\n', poissonfile);
%
fclose(hfile);
%
% ditto mid-res
hfile=fopen(lacqheaderfilem,'w');
fprintf(hfile,'nx=%d\n', nx);
fprintf(hfile,'ny=%d\n', ny);
fprintf(hfile,'nz=%d\n', 1);
fprintf(hfile,'x0=%f\n', x0);
fprintf(hfile,'y0=%f\n', y0);
fprintf(hfile,'z0=%f\n', z0);
fprintf(hfile,'dx=%f\n', dx);
fprintf(hfile,'dy=%f\n', dy);
fprintf(hfile,'dz=%f\n', dz);
%
fprintf(hfile,'rnx=%d\n', rnxm);
fprintf(hfile,'rny=%d\n', rnym);
fprintf(hfile,'rx0=%f\n', rx0);
fprintf(hfile,'ry0=%f\n', ry0);
fprintf(hfile,'rdx=%f\n', rdxm);
fprintf(hfile,'rdy=%f\n', rdym);
%
fprintf(hfile,'real rzlo(rnx,rny)=%s\n', topsurffilem);
fprintf(hfile,'real rzhi(rnx,rny)=%s\n', bottomsurffilem);
fprintf(hfile,'real ksat(nz,nx,ny)=%s\n', ksatfile);
fprintf(hfile,'real kmin(nz,nx,ny)=%s\n', kminfile);
fprintf(hfile,'real mu(nz,nx,ny)=%s\n', shearfile);
fprintf(hfile,'real nu(nz,nx,ny)=%s\n', poissonfile);
%
fclose(hfile);
% ditto high-res
hfile=fopen(lacqheaderfileh,'w');
fprintf(hfile,'nx=%d\n', nx);
fprintf(hfile,'ny=%d\n', ny);
fprintf(hfile,'nz=%d\n', 1);
fprintf(hfile,'x0=%f\n', x0);
fprintf(hfile,'y0=%f\n', y0);
fprintf(hfile,'z0=%f\n', z0);
fprintf(hfile,'dx=%f\n', dx);
fprintf(hfile,'dy=%f\n', dy);
fprintf(hfile,'dz=%f\n', dz);
%
fprintf(hfile,'rnx=%d\n', rnxh);
fprintf(hfile,'rny=%d\n', rnyh);
fprintf(hfile,'rx0=%f\n', rx0);
fprintf(hfile,'ry0=%f\n', ry0);
fprintf(hfile,'rdx=%f\n', rdxh);
fprintf(hfile,'rdy=%f\n', rdyh);
%
fprintf(hfile,'real rzlo(rnx,rny)=%s\n', topsurffileh);
fprintf(hfile,'real rzhi(rnx,rny)=%s\n', bottomsurffileh);
fprintf(hfile,'real ksat(nz,nx,ny)=%s\n', ksatfile);
fprintf(hfile,'real kmin(nz,nx,ny)=%s\n', kminfile);
fprintf(hfile,'real mu(nz,nx,ny)=%s\n', shearfile);
fprintf(hfile,'real nu(nz,nx,ny)=%s\n', poissonfile);
%
fclose(hfile);
%
% model plotting grid
[ccx, ccy]=meshgrid(linspace(x0,x0+(nx-1)*dx,nx),...
    linspace(y0,y0+(ny-1)*dy,ny));
%
figure
    surf(ccx, ccy, zlacq_real);
    xlabel('x - km');
    ylabel('y - km');
    zlabel('m');
    set(get(gca,'XLabel'),'FontSize',14);
    set(get(gca,'YLabel'),'FontSize',14);
    set(get(gca,'ZLabel'),'FontSize',14);
    set(get(gca,'Title'),'FontSize',14);
    set(gca,'ZDir','reverse');
    title('Presumed Vert Displacement at Lacq','FontSize',20);
    shading interp;
    hco=colorbar;
    ylab=get(hco,'YLabel');
    set(ylab,'FontSIze',14);
    set(ylab,'String','subsidence (m)')
    
figure
    [C,h]=contour(ccx, ccy, zlacq_real);
    clabel(C,h);
    xlabel('x - km');
    ylabel('y - km');
    zlabel('m');
    set(gca,'ZDir','reverse');
    title('Presumed Vert Displacement at Lacq in metres');
    shading interp;
    hco=colorbar;
    ylab=get(hco,'YLabel');
    set(ylab,'FontSIze',14);
    set(ylab,'String','subsidence (m)')
    
%
% reservoir plotting grid
[cx, cy]=meshgrid(linspace(rx0,rx0+rdx*(rnx-1),rnx),...
    linspace(ry0,ry0+rdy*(rnx-1),rny));
  
figure
    surf(cx,cy,rzlo);
    set(gca,'ZDir','reverse');
    xlabel('x - km');
    ylabel('y - km');
    zlabel('km');
    set(get(gca,'XLabel'),'FontSize',14);
    set(get(gca,'YLabel'),'FontSize',14);
    set(get(gca,'ZLabel'),'FontSize',14);
    set(get(gca,'Title'),'FontSize',14);    
    title('Reservoir Dome');
    shading interp;
    hco=colorbar;
    ylab=get(hco,'YLabel');
    set(ylab,'FontSIze',14);
    set(ylab,'String','Depth (km)');
